# -*- coding: utf-8 -*-

import sys 
import os.path
import inspect
import ast
import _ast
import _io
import traceback
import types
import logging
import pydoc
import builtins
import site

import __main__  # @UnresolvedImport

from thonny import ast_utils
from thonny.common import TextRange,\
    parse_message, serialize_message, DebuggerCommand,\
    ToplevelCommand, FrameInfo, InlineCommand, InputSubmission
import signal
import warnings

BEFORE_STATEMENT_MARKER = "_thonny_hidden_before_stmt"
BEFORE_EXPRESSION_MARKER = "_thonny_hidden_before_expr"
AFTER_STATEMENT_MARKER = "_thonny_hidden_after_stmt"
AFTER_EXPRESSION_MARKER = "_thonny_hidden_after_expr"

EXCEPTION_TRACEBACK_LIMIT = 100
DEBUG = True    

logger = logging.getLogger()
info = logger.info

class VM:
    def __init__(self):
        self._main_dir = os.path.dirname(sys.modules["thonny"].__file__)
        self._heap = {} # WeakValueDictionary would be better, but can't store reference to None
        site.sethelper() # otherwise help function is not available
        pydoc.pager = pydoc.plainpager # otherwise help command plays tricks
        self._install_fake_streams()
        self._current_executor = None
        self._io_level = 0
        
        original_argv = sys.argv.copy()
        original_path = sys.path.copy()
        
        # clean up path
        sys.path = [d for d in sys.path if d != ""]
        
        # script mode
        if len(sys.argv) > 1:
            special_names_to_remove = set()
            sys.argv[:] = sys.argv[1:] # shift argv[1] to position of script name
            sys.path.insert(0, os.path.abspath(os.path.dirname(sys.argv[0]))) # add program's dir
            __main__.__dict__["__file__"] = sys.argv[0]
            # TODO: inspect.getdoc
        
        # shell mode
        else:
            special_names_to_remove = {"__file__", "__cached__"}
            sys.argv[:] = [""] # empty "script name"
            sys.path.insert(0, "")   # current dir
        
        # add jedi
        if "JEDI_LOCATION" in os.environ:
            sys.path.append(os.environ["JEDI_LOCATION"])
    
        # clean __main__ global scope
        for key in list(__main__.__dict__.keys()):
            if not key.startswith("__") or key in special_names_to_remove:
                del __main__.__dict__[key] 
        
        # unset __doc__, then exec dares to write doc of the script there
        __main__.__doc__ = None
        
        self.send_message(self.create_message("ToplevelResult",
                          main_dir=self._main_dir,
                          original_argv=original_argv,
                          original_path=original_path,
                          argv=sys.argv,
                          path=sys.path,
                          welcome_text="Python " + _get_python_version_string(),
                          executable=sys.executable,
                          in_venv=hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix,
                          python_version=_get_python_version_string(),
                          cwd=os.getcwd()))
        
        self._install_signal_handler()
        
    def mainloop(self):
        try:
            while True: 
                try:
                    cmd = self._fetch_command()
                    self.handle_command(cmd, "waiting_toplevel_command")
                except KeyboardInterrupt:
                    logger.exception("Interrupt in mainloop")
                    # Interrupt must always result in waiting_toplevel_command state
                    # Don't show error messages, as the interrupted command may have been InlineCommand
                    # (handlers of ToplevelCommands in normal cases catch the interrupt and provide
                    # relevant message)  
                    self.send_message(self.create_message("ToplevelResult"))
        except:
            logger.exception("Crash in mainloop")
            
            
    def handle_command(self, cmd, command_context):
        assert isinstance(cmd, ToplevelCommand) or isinstance(cmd, InlineCommand)
        
        error_response_type = "ToplevelResult" if isinstance(cmd, ToplevelCommand) else "InlineError"
        try:
            handler = getattr(self, "_cmd_" + cmd.command)
        except AttributeError:
            response = self.create_message(error_response_type, error="Unknown command: " + cmd.command)
        else:
            try:
                response = handler(cmd)
            except:
                response = self.create_message(error_response_type,
                    error="Thonny internal error: {0}".format(traceback.format_exc(EXCEPTION_TRACEBACK_LIMIT)))
        
        if response is not None:
            response["command_context"] = command_context
            response["command"] = cmd.command
            if response["message_type"] == "ToplevelResult":
                self._add_tkinter_info(response)
            self.send_message(response)
    
    def _install_signal_handler(self):
        def signal_handler(signal, frame):
            raise KeyboardInterrupt("Execution interrupted")
        
        if os.name == 'nt':
            signal.signal(signal.SIGBREAK, signal_handler)
        else:
            signal.signal(signal.SIGINT, signal_handler)        
    
    def _cmd_cd(self, cmd):
        try:
            os.chdir(cmd.path)
            return self.create_message("ToplevelResult")
        except Exception as e:
            # TODO: should output user error
            return self.create_message("ToplevelResult", error=str(e))
    
    def _cmd_Reset(self, cmd):
        # nothing to do, because Reset always happens in fresh process
        return self.create_message("ToplevelResult",
                                   welcome_text="Python " + _get_python_version_string(),
                                   executable=sys.executable)
    
    def _cmd_Run(self, cmd):
        return self._execute_file(cmd, False)
    
    def _cmd_run(self, cmd):
        return self._execute_file(cmd, False)
    
    def _cmd_Debug(self, cmd):
        return self._execute_file(cmd, True)
    
    def _cmd_debug(self, cmd):
        return self._execute_file(cmd, True)
    
    def _cmd_execute_source(self, cmd):
        return self._execute_source(cmd, "ToplevelResult")
    
    def _cmd_execute_source_inline(self, cmd):
        return self._execute_source(cmd, "InlineResult")
    
    def _cmd_tkupdate(self, cmd):
        # advance the event loop
        # http://bugs.python.org/issue989712
        # http://bugs.python.org/file6090/run.py.diff
        try:
            root = self._get_tkinter_default_root()
            if root is None:
                return
            
            import tkinter
            while root.dooneevent(tkinter._tkinter.DONT_WAIT):
                pass
                 
        except:
            pass
            
        return None
    
    
    def _cmd_get_globals(self, cmd):
        if not cmd.module_name in sys.modules:
            raise ThonnyClientError("Module '{0}' is not loaded".format(cmd.module_name))
        
        return self.create_message("Globals", module_name=cmd.module_name,
                              globals=self.export_variables(sys.modules[cmd.module_name].__dict__))
    
    def _cmd_get_locals(self, cmd):
        for frame in inspect.stack():
            if id(frame) == cmd.frame_id:
                return self.create_message("Locals", locals=self.export_variables(frame.f_locals))
        else:
            raise ThonnyClientError("Frame '{0}' not found".format(cmd.frame_id))
            
    
    def _cmd_get_heap(self, cmd):
        result = {}
        for key in self._heap:
            result[key] = self.export_value(self._heap[key])
            
        return self.create_message("Heap", heap=result)
    
    def _cmd_shell_autocomplete(self, cmd):
        error = None
        try:
            import jedi
        except ImportError:
            completions = []
            error = "Could not import jedi"
        else:
            try:
                #with warnings.catch_warnings():
                interpreter = jedi.Interpreter(cmd.source, [__main__.__dict__])
                completions = self._export_completions(interpreter.completions())
            except Exception as e:
                completions = []
                error = "Autocomplete error: " + str(e)
            except:
                completions = []
                error = "Autocomplete error"
        
        return self.create_message("ShellCompletions", 
            source=cmd.source,
            completions=completions,
            error=error
        )
    
    def _cmd_editor_autocomplete(self, cmd):
        error = None
        try:
            import jedi
            with warnings.catch_warnings():
                script = jedi.Script(cmd.source, cmd.row, cmd.column, cmd.filename)
                completions = self._export_completions(script.completions())
                
        except ImportError:
            completions = []
            error = "Could not import jedi"
        except Exception as e:
            completions = []
            error = "Autocomplete error: " + str(e)
        except:
            completions = []
            error = "Autocomplete error"
        
        return self.create_message("EditorCompletions", 
                          source=cmd.source,
                          row=cmd.row,
                          column=cmd.column,
                          filename=cmd.filename,
                          completions=completions,
                          error=error)
    
    def _export_completions(self, jedi_completions):
        result = []
        for c in jedi_completions:
            if not c.name.startswith("__"):
                record = {"name":c.name, "complete":c.complete, 
                          "type":c.type, "description":c.description}
                try:
                    """ TODO: 
                    if c.type in ["class", "module", "function"]:
                        if c.type == "function":
                            record["docstring"] = c.docstring()
                        else:
                            record["docstring"] = c.description + "\n" + c.docstring()
                    """
                except:
                    pass
                result.append(record)
        return result
        
    
    def _cmd_get_object_info(self, cmd):
        if cmd.object_id in self._heap:
            value = self._heap[cmd.object_id]
            attributes = {}
            for name in dir(value):
                if not name.startswith("__") or cmd.all_attributes:
                    #attributes[name] = inspect.getattr_static(value, name)
                    try: 
                        attributes[name] = getattr(value, name)
                    except:
                        pass 
            
            self._heap[id(type(value))] = type(value)
            
            info = {'id' : cmd.object_id,
                    'repr' : repr(value),
                    'type' : str(type(value)),
                    'type_id' : id(type(value)),
                    'attributes': self.export_variables(attributes)}
            
            if isinstance(value, _io.TextIOWrapper):
                self._add_file_handler_info(value, info)
            elif (type(value) in (types.BuiltinFunctionType, types.BuiltinMethodType,
                                 types.FunctionType, types.LambdaType, types.MethodType)):
                self._add_function_info(value, info)
            elif (isinstance(value, list) 
                  or isinstance(value, tuple)
                  or isinstance(value, set)):
                self._add_elements_info(value, info)
            elif (isinstance(value, dict)):
                self._add_entries_info(value, info)
            
        else:
            info = {'id' : cmd.object_id,
                    "repr": "<object info not found>",
                    "type" : "object",
                    "type_id" : id(object),
                    "attributes" : {}}
        
        return self.create_message("ObjectInfo", id=cmd.object_id, info=info)
    
    def _get_tkinter_default_root(self):
        tkinter = sys.modules.get("tkinter")
        if tkinter is not None:
            return getattr(tkinter, "_default_root", None)
        else:
            return None

    
    def _add_file_handler_info(self, value, info):
        try:
            assert isinstance(value.name, str)
            assert value.mode in ("r", "rt", "tr", "br", "rb")
            assert value.errors in ("strict", None)
            assert value.newlines is None or value.tell() > 0
            # TODO: cache the content
            # TODO: don't read too big files
            with open(value.name, encoding=value.encoding) as f:
                info["file_encoding"] = f.encoding
                info["file_content"] = f.read()
                info["file_tell"] = value.tell()
        except Exception as e:
            info["file_error"] = "Could not get file content, error:" + str(e)
            pass
    
    def _add_tkinter_info(self, msg):
        # tkinter._default_root is not None,
        # when window has been created and mainloop isn't called or hasn't ended yet
        msg["tkinter_is_active"] = self._get_tkinter_default_root() is not None
    
    def _add_function_info(self, value, info):
        try:
            info["source"] = inspect.getsource(value)
        except:
            pass
        
    def _add_elements_info(self, value, info):
        info["elements"] = []
        for element in value:
            info["elements"].append(self.export_value(element))
        
    def _add_entries_info(self, value, info):
        info["entries"] = []
        for key in value:
            info["entries"].append((self.export_value(key),
                                     self.export_value(value[key])))
        
    def _execute_file(self, cmd, debug_mode):
        # args are accepted only in Run and Debug,
        # and were stored in sys.argv already in VM.__init__
        result_attributes = self._execute_source_ex(cmd.source, cmd.full_filename, "exec", debug_mode) 
        return self.create_message("ToplevelResult", **result_attributes)
    
    def _execute_source(self, cmd, result_type):
        filename = "<pyshell>"
        
        if hasattr(cmd, "global_vars"):
            global_vars = cmd.global_vars
        elif hasattr(cmd, "extra_vars"):
            global_vars = __main__.__dict__.copy() # Don't want to mess with main namespace
            global_vars.update(cmd.extra_vars)
        else:
            global_vars = __main__.__dict__

        # let's see if it's single expression or something more complex
        try:
            root = ast.parse(cmd.source, filename=filename, mode="exec")
        except SyntaxError as e:
            return self.create_message(result_type,
                error="".join(traceback.format_exception_only(SyntaxError, e)))
            
        assert isinstance(root, ast.Module)
            
        if len(root.body) == 1 and isinstance(root.body[0], ast.Expr):
            mode = "eval"
        else:
            mode = "exec"
            
        result_attributes = self._execute_source_ex(cmd.source, filename, mode,
            hasattr(cmd, "debug_mode") and cmd.debug_mode,
            global_vars)
        
        if "__result__" in global_vars:
            result_attributes["__result__"] = global_vars["__result__"]
        
        if hasattr(cmd, "request_id"):
            result_attributes["request_id"] = cmd.request_id
        else:
            result_attributes["request_id"] = None
        
        return self.create_message(result_type, **result_attributes)
        
    def _execute_source_ex(self, source, filename, execution_mode, debug_mode,
                        global_vars=None):
        if debug_mode:
            self._current_executor = FancyTracer(self)
        else:
            self._current_executor = Executor(self)
        
        try:
            return self._current_executor.execute_source(source, 
                                                         filename,
                                                         execution_mode,
                                                         global_vars)
        finally:
            self._current_executor = None
    
        
    def _install_fake_streams(self):
        self._original_stdin = sys.stdin
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr        
        
        # yes, both out and err will be directed to out (but with different tags)
        # this allows client to see the order of interleaving writes to stdout/stderr
        sys.stdin = VM.FakeInputStream(self, sys.stdin)
        sys.stdout = VM.FakeOutputStream(self, sys.stdout, "stdout")
        sys.stderr = VM.FakeOutputStream(self, sys.stdout, "stderr") 
             
        # fake it properly: replace also "backup" streams
        sys.__stdin__ = sys.stdin
        sys.__stdout__ = sys.stdout
        sys.__stderr__ = sys.stderr
        
    def _fetch_command(self):
        line = self._original_stdin.readline()
        if line == "":
            logger.info("Read stdin EOF")
            sys.exit()
        cmd = parse_message(line)
        return cmd

    def create_message(self, message_type, **kwargs):
        kwargs["message_type"] = message_type
        if "cwd" not in kwargs:
            kwargs["cwd"] = os.getcwd()
            
        return kwargs

    def send_message(self, msg):
        self._original_stdout.write(serialize_message(msg) + "\n")
        self._original_stdout.flush()
        
    def export_value(self, value, skip_None=False):
        if value is None and skip_None:
            return None
        
        self._heap[id(value)] = value
        try:
            type_name = value.__class__.__name__
        except:
            type_name = type(value).__name__ 
            
        result = {'id' : id(value),
                  'repr' : repr(value), 
                  'type_name'  : type_name}
        
        return result
    
    def export_variables(self, variables):
        result = {}
        for name in variables:
            if not name.startswith("_thonny_hidden_"):
                result[name] = self.export_value(variables[name])
            
        return result
           
    def _debug(self, *args):
        print("VM:", *args, file=self._original_stderr)
    
    
    def _enter_io_function(self):
        self._io_level += 1
    
    def _exit_io_function(self):
        self._io_level -= 1
    
    def is_doing_io(self):
        return self._io_level > 0
    

    class FakeStream:
        def __init__(self, vm, target_stream):
            self._vm = vm
            self._target_stream = target_stream
            
        def isatty(self):
            return True
        
        def __getattr__(self, name):
            # TODO: is it safe to perform those other functions without notifying vm
            # via _enter_io_function?
            return getattr(self._target_stream, name)
        
    class FakeOutputStream(FakeStream):
        def __init__(self, vm, target_stream, stream_name):
            VM.FakeStream.__init__(self, vm, target_stream)
            self._stream_name = stream_name
            
        def write(self, data):
            try:
                self._vm._enter_io_function()
                if data != "":
                    self._vm.send_message(self._vm.create_message("ProgramOutput", stream_name=self._stream_name, data=data))
            finally:
                self._vm._exit_io_function()
        
        def writelines(self, lines):
            try:
                self._vm._enter_io_function()
                self.write(''.join(lines))
            finally:
                self._vm._exit_io_function()
    
    class FakeInputStream(FakeStream):
        
        def _generic_read(self, method, limit=-1):
            try:
                self._vm._enter_io_function()
                self._vm.send_message(self._vm.create_message("InputRequest", method=method, limit=limit))
                
                while True:
                    cmd = self._vm._fetch_command()
                    if isinstance(cmd, InputSubmission):
                        return cmd.data
                    elif isinstance(cmd, InlineCommand):
                        self._vm.handle_command(cmd, "waiting_input")
                    else:
                        raise ThonnyClientError("Wrong type of command when waiting for input")
            finally:
                self._vm._exit_io_function()
        
        def read(self, limit=-1):
            return self._generic_read("read", limit)
        
        def readline(self, limit=-1):
            return self._generic_read("readline", limit)
        
        def readlines(self, limit=-1):
            return self._generic_read("readlines", limit)
            
    


class Executor:
    def __init__(self, vm):
        self._vm = vm
    
    def execute_source(self, source, filename, mode, global_vars=None):
        
        if global_vars is None:
            global_vars = __main__.__dict__
        
        try:
            bytecode = self._compile_source(source, filename, mode)
            if hasattr(self, "_trace"):
                sys.settrace(self._trace)    
            if mode == "eval":
                value = eval(bytecode, global_vars)
                if value is not None:
                    builtins._ = value 
                return {"value_info" : self._vm.export_value(value)}
            else:
                assert mode == "exec"
                exec(bytecode, global_vars) # <Marker: remove this line from stacktrace>
                return {"context_info" : "after normal execution", "source" : source, "filename" : filename, "mode" : mode}
        except SyntaxError as e:
            return {"error" : "".join(traceback.format_exception_only(SyntaxError, e))}
        except ThonnyClientError as e:
            return {"error" : str(e)}
        except SystemExit:
            e_type, e_value, e_traceback = sys.exc_info()
            self._print_user_exception(e_type, e_value, e_traceback)
            return {"SystemExit" : True}
        except:
            # other unhandled exceptions (supposedly client program errors) are printed to stderr, as usual
            # for VM mainloop they are not exceptions
            e_type, e_value, e_traceback = sys.exc_info()
            self._print_user_exception(e_type, e_value, e_traceback)
            return {"context_info" : "other unhandled exception"}
        finally:
            sys.settrace(None)
    
    def _print_user_exception(self, e_type, e_value, e_traceback):
        lines = traceback.format_exception(e_type, e_value, e_traceback)

        for line in lines:
            # skip lines denoting thonny execution frame
            if ("thonny/backend" in line 
                or "thonny\\backend" in line
                or "remove this line from stacktrace" in line):
                continue
            else:
                sys.stderr.write(line)

    def _compile_source(self, source, filename, mode):
        return compile(source, filename, mode)


class FancyTracer(Executor):
    
    def __init__(self, vm):
        self._vm = vm
        self._normcase_thonny_src_dir = os.path.normcase(os.path.dirname(sys.modules["thonny"].__file__)) 
        self._instrumented_files = _PathSet()
        self._interesting_files = _PathSet() # only events happening in these files are reported
        self._current_command = None
        self._unhandled_exception = None
        self._install_marker_functions()
        self._custom_stack = []
    
    def execute_source(self, source, filename, mode, global_vars=None):
        self._current_command = DebuggerCommand(command="step", state=None, focus=None, frame_id=None, exception=None)
        
        return Executor.execute_source(self, source, filename, mode, global_vars)
        #assert len(self._custom_stack) == 0
        
    def _install_marker_functions(self):
        # Make dummy marker functions universally available by putting them
        # into builtin scope        
        self.marker_function_names = {
            BEFORE_STATEMENT_MARKER,
            AFTER_STATEMENT_MARKER, 
            BEFORE_EXPRESSION_MARKER,
            AFTER_EXPRESSION_MARKER,
        }
        
        for name in self.marker_function_names:
            if not hasattr(builtins, name):
                setattr(builtins, name, getattr(self, name))
        
    def _is_interesting_exception(self, frame):
        # interested only in exceptions in command frame or it's parent frames
        cmd = self._current_command
        return (id(frame) == cmd.frame_id
                or not self._frame_is_alive(cmd.frame_id))

    def _compile_source(self, source, filename, mode):
        root = ast.parse(source, filename, mode)
        
        ast_utils.mark_text_ranges(root, source)
        self._tag_nodes(root)        
        self._insert_expression_markers(root)
        self._insert_statement_markers(root)
        self._instrumented_files.add(filename)
        
        return compile(root, filename, mode)
    
    def _may_step_in(self, code):
            
        return not (
            code is None 
            or code.co_filename is None
            or code.co_flags & inspect.CO_GENERATOR  # @UndefinedVariable
            or sys.version_info >= (3,5) and code.co_flags & inspect.CO_COROUTINE  # @UndefinedVariable
            or sys.version_info >= (3,5) and code.co_flags & inspect.CO_ITERABLE_COROUTINE  # @UndefinedVariable
            or sys.version_info >= (3,6) and code.co_flags & inspect.CO_ASYNC_GENERATOR  # @UndefinedVariable
            or "importlib._bootstrap" in code.co_filename
            or os.path.normcase(code.co_filename) not in self._instrumented_files 
                and code.co_name not in self.marker_function_names
            or os.path.normcase(code.co_filename).startswith(self._normcase_thonny_src_dir)
                and code.co_name not in self.marker_function_names
            or self._vm.is_doing_io() 
        )
        
    
    def _trace(self, frame, event, arg):
        """
        1) Detects marker calls and responds to client queries in these spots
        2) Maintains a customized view of stack
        """
        if not self._may_step_in(frame.f_code):
            return
        
        code_name = frame.f_code.co_name
        
        if event == "call":
            self._unhandled_exception = None # some code is running, therefore exception is not propagating anymore
            
            if code_name in self.marker_function_names:
                # the main thing
                if code_name == BEFORE_STATEMENT_MARKER:
                    event = "before_statement"
                elif code_name == AFTER_STATEMENT_MARKER:
                    event = "after_statement"
                elif code_name == BEFORE_EXPRESSION_MARKER:
                    event = "before_expression"
                elif code_name == AFTER_EXPRESSION_MARKER:
                    event = "after_expression"
                else:
                    raise AssertionError("Unknown marker function")
                
                marker_function_args = frame.f_locals.copy()
                del marker_function_args["self"]
                
                self._handle_progress_event(frame.f_back, event, marker_function_args)
                self._try_interpret_as_again_event(frame.f_back, event, marker_function_args)
                
                
            else:
                # Calls to proper functions.
                # Client doesn't care about these events,
                # it cares about "before_statement" events in the first statement of the body
                self._custom_stack.append(CustomStackFrame(frame, "call"))
        
        elif event == "return":
            if code_name not in self.marker_function_names:
                self._custom_stack.pop()
                if len(self._custom_stack) == 0:
                    # We popped last frame, this means our program has ended.
                    # There may be more events coming from upper (system) frames
                    # but we're not interested in those
                    sys.settrace(None)
            else:
                pass
                
        elif event == "exception":
            exc = arg[1]
            if self._unhandled_exception is None:
                # this means it's the first time we see this exception
                exc.causing_frame = frame
            else:
                # this means the exception is propagating to older frames
                # get the causing_frame from previous occurrence
                exc.causing_frame = self._unhandled_exception.causing_frame 
            
            self._unhandled_exception = exc
            if self._is_interesting_exception(frame):
                self._report_state_and_fetch_next_message(frame)

        # TODO: support line event in non-instrumented files
        elif event == "line":
            self._unhandled_exception = None  
                
        return self._trace
        
            
    def _handle_progress_event(self, frame, event, args):
        """
        Tries to respond to current command in this state. 
        If it can't, then it returns, program resumes
        and _trace will call it again in another state.
        Otherwise sends response and fetches next command.  
        """
        self._debug("Progress event:", event, self._current_command)
        focus = TextRange(*args["text_range"])
        
        self._custom_stack[-1].last_event = event
        self._custom_stack[-1].last_event_focus = focus
        self._custom_stack[-1].last_event_args = args
        
        # Select the correct method according to the command
        tester = getattr(self, "_cmd_" + self._current_command.command + "_completed")
             
        # If method decides we're in the right place to respond to the command ...
        if tester(frame, event, args, focus, self._current_command):
            if event == "after_expression":
                value = self._vm.export_value(args["value"])
            else:
                value = None
            self._report_state_and_fetch_next_message(frame, value)
    
    def _report_state_and_fetch_next_message(self, frame, value=None):
            #self._debug("Completed command: ", self._current_command)
            
            if self._unhandled_exception is not None:
                frame_infos = traceback.format_stack(self._unhandled_exception.causing_frame)
                # I want to show frames from current frame to causing_frame
                if frame == self._unhandled_exception.causing_frame:
                    interesting_frame_infos = []
                else:
                    # c how far is current frame from causing_frame?
                    _distance = 0
                    _f = self._unhandled_exception.causing_frame 
                    while _f != frame:
                        _distance += 1
                        _f = _f.f_back
                        if _f == None:
                            break
                    interesting_frame_infos = frame_infos[-_distance:]
                exception_lower_stack_description = "".join(interesting_frame_infos)
                exception_msg = str(self._unhandled_exception)
            else:
                exception_lower_stack_description = None 
                exception_msg = None
            
            self._vm.send_message(self._vm.create_message("DebuggerProgress",
                command=self._current_command.command,
                stack=self._export_stack(),
                exception=self._vm.export_value(self._unhandled_exception, True),
                exception_msg=exception_msg,
                exception_lower_stack_description=exception_lower_stack_description,
                value=value,
                command_context="waiting_debugger_command"
            ))
            
            # Fetch next debugger command
            self._current_command = self._vm._fetch_command()
            self._debug("got command:", self._current_command)
            # get non-progress commands out our way
            self._respond_to_inline_commands()  
            assert isinstance(self._current_command, DebuggerCommand)
            
        # Return and let Python run to next progress event
        
    
    def _try_interpret_as_again_event(self, frame, original_event, original_args):
        """
        Some after_* events can be interpreted also as 
        "before_*_again" events (eg. when last argument of a call was 
        evaluated, then we are just before executing the final stage of the call)
        """

        if original_event == "after_expression":
            node_tags = original_args.get("node_tags")
            value = original_args.get("value")
            
            if (node_tags is not None 
                and ("last_child" in node_tags
                     or "or_arg" in node_tags and value
                     or "and_arg" in node_tags and not value)):
                
                # next step will be finalizing evaluation of parent of current expr
                # so let's say we're before that parent expression
                again_args = {"text_range" : original_args.get("parent_range"),
                              "node_tags" : ""}
                again_event = ("before_expression_again" 
                               if "child_of_expression" in node_tags
                               else "before_statement_again")
                
                self._handle_progress_event(frame, again_event, again_args)
                
    
    def _respond_to_inline_commands(self):
        while isinstance(self._current_command, InlineCommand): 
            self._vm.handle_command(self._current_command, "waiting_debugger_command")
            self._current_command = self._vm._fetch_command()
    
    def _get_frame_source_info(self, frame):
        if frame.f_code.co_name == "<module>":
            obj = inspect.getmodule(frame)
            lineno = 1
        else:
            obj = frame.f_code
            lineno = obj.co_firstlineno
        
        # lineno returned by getsourcelines is not consistent between modules vs functions
        lines, _ = inspect.getsourcelines(obj) 
        return "".join(lines), lineno
        
    
    
    def _cmd_exec_completed(self, frame, event, args, focus, cmd):
        """
        Identifies the moment when piece of code indicated by cmd.frame_id and cmd.focus
        has completed execution (either successfully or not).
        """
        
        # it's meant to be executed in before* state, but if we're not there
        # we'll step there
        
        if cmd.state not in ("before_expression", "before_expression_again",
                             "before_statement", "before_statement_again"):
            return self._cmd_step_completed(frame, event, args, focus, cmd)
        
        
        if id(frame) == cmd.frame_id:
            
            if focus.is_smaller_in(cmd.focus):
                # we're executing a child of command focus,
                # keep running
                return False 
            
            elif focus == cmd.focus:
                
                if event.startswith("before_"):
                    # we're just starting
                    return False
                
                elif (event == "after_expression"
                      and cmd.state in ("before_expression", "before_expression_again")
                      or 
                      event == "after_statement"
                      and cmd.state in ("before_statement", "before_statement_again")):
                    # Normal completion
                    # Maybe there was an exception, but this is forgotten now
                    cmd._unhandled_exception = False
                    self._debug("Exec normal")
                    return True
                
                
                elif (cmd.state in ("before_statement", "before_statement_again")
                      and event == "after_expression"):
                    # Same code range can contain expression statement and expression.
                    # Here we need to run just a bit more
                    return False
                
                else:
                    # shouldn't be here
                    raise AssertionError("Unexpected state in responding to " + str(cmd))
                    
            else:
                # We're outside of starting focus, assumedly because of an exception
                self._debug("Exec outside", cmd.focus, focus)
                return True
        
        else:
            # We're in another frame
            if self._frame_is_alive(cmd.frame_id):
                # We're in a successor frame, keep running
                return False
            else:
                # Original frame has completed, assumedly because of an exception
                # We're done
                self._debug("Exec wrong frame")
                return True
            

    
    def _cmd_step_completed(self, frame, event, args, focus, cmd):
        return True
    
    def _cmd_run_to_before_completed(self, frame, event, args, focus, cmd):
        return event.startswith("before")
    
    def _cmd_out_completed(self, frame, event, args, focus, cmd):
        """Complete current frame"""
        return (
            # the frame has completed
            not self._frame_is_alive(cmd.frame_id)
            # we're in the same frame but on higher level 
            or id(frame) == cmd.frame_id and focus.contains_smaller(cmd.focus)
        )
    
    
    def _cmd_line_completed(self, frame, event, args, focus, cmd):
        return (event == "before_statement" 
            and os.path.normcase(frame.f_code.co_filename) == os.path.normcase(cmd.target_filename)
            and focus.lineno == cmd.target_lineno
            and (focus != cmd.focus or id(frame) != cmd.frame_id))

    
    def _frame_is_alive(self, frame_id):
        for frame in self._custom_stack:
            if frame.id == frame_id:
                return True
        else:
            return False 
    
    def _export_stack(self):
        result = []
        
        for custom_frame in self._custom_stack:
            
            last_event_args = custom_frame.last_event_args.copy()
            if "value" in last_event_args:
                last_event_args["value"] = self._vm.export_value(last_event_args["value"]) 
            
            system_frame = custom_frame.system_frame
            source, firstlineno = self._get_frame_source_info(system_frame)
            
            result.append(FrameInfo(
                id=id(system_frame),
                filename=system_frame.f_code.co_filename,
                module_name=system_frame.f_globals["__name__"],
                code_name=system_frame.f_code.co_name,
                locals=self._vm.export_variables(system_frame.f_locals),
                source=source,
                firstlineno=firstlineno,
                last_event=custom_frame.last_event,
                last_event_args=last_event_args,
                last_event_focus=custom_frame.last_event_focus,
            ))
        
        return result

    def _thonny_hidden_before_stmt(self, text_range, node_tags):
        """
        The code to be debugged will be instrumented with this function
        inserted before each statement. 
        Entry into this function indicates that statement as given
        by the code range is about to be evaluated next.
        """
        return None
    
    def _thonny_hidden_after_stmt(self, text_range, node_tags):
        """
        The code to be debugged will be instrumented with this function
        inserted after each statement. 
        Entry into this function indicates that statement as given
        by the code range was just executed successfully.
        """
        return None
    
    def _thonny_hidden_before_expr(self, text_range, node_tags):
        """
        Entry into this function indicates that expression as given
        by the code range is about to be evaluated next
        """ 
        return text_range
    
    def _thonny_hidden_after_expr(self, text_range, node_tags, value, parent_range):
        """
        The code to be debugged will be instrumented with this function
        wrapped around each expression (given as 2nd argument). 
        Entry into this function indicates that expression as given
        by the code range was just evaluated to given value
        """ 
        return value
    

    def _tag_nodes(self, root):
        """Marks interesting properties of AST nodes"""
        
        def add_tag(node, tag):
            if not hasattr(node, "tags"):
                node.tags = set()
                node.tags.add("class=" + node.__class__.__name__)
            node.tags.add(tag)
        
        for node in ast.walk(root):
            
            # tag last children 
            last_child = ast_utils.get_last_child(node)
            if last_child is not None and last_child:
                add_tag(node, "has_children")
                
                if isinstance(last_child, ast.AST):
                    last_child.parent_node = node
                    add_tag(last_child, "last_child")
                    if isinstance(node, _ast.expr):
                        add_tag(last_child, "child_of_expression")
                    else:
                        add_tag(last_child, "child_of_statement")
                    
                    if isinstance(node, ast.Call):
                        add_tag(last_child, "last_call_arg")
                    
            # other cases
            if isinstance(node, ast.Call):
                add_tag(node.func, "call_function")
                node.func.parent_node = node
                
            if isinstance(node, ast.BoolOp) and node.op == ast.Or():
                for child in node.values:
                    add_tag(child, "or_arg")
                    child.parent_node = node
                    
            if isinstance(node, ast.BoolOp) and node.op == ast.And():
                for child in node.values:
                    add_tag(child, "and_arg")
                    child.parent_node = node
            
            # TODO: assert (it doesn't evaluate msg when test == True)
            
            if isinstance(node, ast.Str):
                add_tag(node, "StringLiteral")
                
            if isinstance(node, ast.Num):
                add_tag(node, "NumberLiteral")
            
            if isinstance(node, ast.ListComp):
                add_tag(node.elt, "ListComp.elt")
                
            if isinstance(node, ast.SetComp):
                add_tag(node.elt, "SetComp.elt")
                
            if isinstance(node, ast.DictComp):
                add_tag(node.key, "DictComp.key")
                add_tag(node.value, "DictComp.value")
            
            if isinstance(node, ast.comprehension):
                for expr in node.ifs:
                    add_tag(expr, "comprehension.if")
                
                
            # make sure every node has this field
            if not hasattr(node, "tags"):
                node.tags = set()
            
    
    def _should_instrument_as_expression(self, node):
        return (isinstance(node, _ast.expr)
                and (not hasattr(node, "ctx") or isinstance(node.ctx, ast.Load))
                # TODO: repeatedly evaluated subexpressions of comprehensions
                # can be supported (but it requires some redesign both in backend and GUI)
                and "ListComp.elt" not in node.tags 
                and "SetComp.elt" not in node.tags
                and "DictComp.key" not in node.tags
                and "DictComp.value" not in node.tags
                and "comprehension.if" not in node.tags
                )         
        return 
    
    def _should_instrument_as_statement(self, node):
        return (isinstance(node, _ast.stmt)
                # Shouldn't insert anything before from __future__ import
                # as this is not a normal statement
                # https://bitbucket.org/plas/thonny/issues/183/thonny-throws-false-positive-syntaxerror
                and (not isinstance(node, _ast.ImportFrom)
                     or node.module != "__future__"))
    
    def _insert_statement_markers(self, root):
        # find lists of statements and insert before/after markers for each statement
        for name, value in ast.iter_fields(root):
            if isinstance(value, ast.AST):
                self._insert_statement_markers(value)
            elif isinstance(value, list):
                if len(value) > 0:
                    new_list = []
                    for node in value:
                        if self._should_instrument_as_statement(node):
                            # self._debug("EBFOMA", node)
                            # add before marker
                            new_list.append(self._create_statement_marker(node, 
                                                                          BEFORE_STATEMENT_MARKER))
                        
                        # original statement
                        if self._should_instrument_as_statement(node):
                            self._insert_statement_markers(node)
                        new_list.append(node)
                        
                        if isinstance(node, _ast.stmt):
                            # add after marker
                            new_list.append(self._create_statement_marker(node,
                                                                          AFTER_STATEMENT_MARKER))
                    setattr(root, name, new_list)
    
    
    def _create_statement_marker(self, node, function_name):
        call = self._create_simple_marker_call(node, function_name)
        stmt = ast.Expr(value=call)
        ast.copy_location(stmt, node)
        ast.fix_missing_locations(stmt)
        return stmt
        
    
    def _insert_expression_markers(self, node):
        """
        each expression e gets wrapped like this:
            _after(_before(_loc, _node_is_zoomable), e, _node_role, _parent_range)
        where
            _after is function that gives the resulting value
            _before is function that signals the beginning of evaluation of e
            _loc gives the code range of e
            _node_is_zoomable indicates whether this node has subexpressions
            _node_role is either 'last_call_arg', 'last_op_arg', 'first_or_arg', 
                                 'first_and_arg', 'function' or None
        """
        tracer = self
        
        class ExpressionVisitor(ast.NodeTransformer):
            def generic_visit(self, node):
                if isinstance(node, _ast.expr):
                    if isinstance(node, ast.Starred):
                        # keep this node as is, but instrument its children
                        return ast.NodeTransformer.generic_visit(self, node)
                    elif tracer._should_instrument_as_expression(node):
                        # before marker 
                        before_marker = tracer._create_simple_marker_call(node, BEFORE_EXPRESSION_MARKER)
                        ast.copy_location(before_marker, node)
                        
                        # after marker
                        after_marker = ast.Call (
                            func=ast.Name(id=AFTER_EXPRESSION_MARKER, ctx=ast.Load()),
                            args=[
                                before_marker,
                                tracer._create_tags_literal(node),
                                ast.NodeTransformer.generic_visit(self, node),
                                tracer._create_location_literal(node.parent_node if hasattr(node, "parent_node") else None)
                            ],
                            keywords=[]
                        )
                        ast.copy_location(after_marker, node)
                        ast.fix_missing_locations(after_marker)
                        
                        return after_marker
                    else:
                        # This expression (and its children) should be ignored
                        return node
                else:
                    # Descend into statements
                    return ast.NodeTransformer.generic_visit(self, node)
        
        return ExpressionVisitor().visit(node)   
            
    
    def _create_location_literal(self, node):
        if node is None:
            return ast_utils.value_to_literal(None)
        
        assert hasattr(node, "end_lineno")
        assert hasattr(node, "end_col_offset")
        
        nums = []
        for value in node.lineno, node.col_offset, node.end_lineno, node.end_col_offset:
            nums.append(ast.Num(n=value))
        return ast.Tuple(elts=nums, ctx=ast.Load())
    
    def _create_tags_literal(self, node):
        if hasattr(node, "tags"):
            # maybe set would perform as well, but I think string is faster
            return ast_utils.value_to_literal(",".join(node.tags))
            #self._debug("YESTAGS")
        else:
            #self._debug("NOTAGS " + str(node))
            return ast_utils.value_to_literal("")
    
    def _create_simple_marker_call(self, node, fun_name):
        assert hasattr(node, "end_lineno")
        assert hasattr(node, "end_col_offset")
        
        args = [
            self._create_location_literal(node),
            self._create_tags_literal(node),
        ]
        
        return ast.Call (
            func=ast.Name(id=fun_name, ctx=ast.Load()),
            args=args,
            keywords=[]
        )
    
    def _debug(self, *args):
        print("TRACER:", *args, file=self._vm._original_stderr)

class CustomStackFrame:
    def __init__(self, frame, last_event, focus=None):
        self.id = id(frame)
        self.system_frame = frame
        self.last_event = last_event
        self.focus = None
        
class ThonnyClientError(Exception):
    pass
    

def fdebug(frame, msg, *args):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(_get_frame_prefix(frame) + msg, *args)
    
    
def _get_frame_prefix(frame):
    return str(id(frame)) + " " + ">" * len(inspect.getouterframes(frame, 0)) + " "
    
def _get_python_version_string(add_word_size=False):
    result = ".".join(map(str, sys.version_info[:3]))
    if sys.version_info[3] != "final":
        result += "-" + sys.version_info[3]
    
    if add_word_size:
        result += " (" + ("64" if sys.maxsize > 2**32 else "32")+ " bit)"
    
    return result    

class _PathSet:
    "implementation of set whose in operator works well for filenames"
    def __init__(self):
        self._normcase_set = set()
        
    def add(self, name):
        self._normcase_set.add(os.path.normcase(name))
    
    def remove(self, name):
        self._normcase_set.remove(os.path.normcase(name))
    
    def clear(self):
        self._normcase_set.clear()
    
    def __contains__(self, name):
        return os.path.normcase(name) in self._normcase_set

    def __iter__(self):
        for item in self._normcase_set:
            yield item



import tkinter as tk
import tkinter.font
import os.path
from tkinter import ttk
from thonny import tktextext
from thonny.globals import get_workbench

class HelpView(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master)
        
        main_font = tkinter.font.nametofont("TkDefaultFont")
        
        bold_font = main_font.copy()
        bold_font.configure(weight="bold", size=main_font.cget("size"))
        
        h1_font = main_font.copy()
        h1_font.configure(size=main_font.cget("size") * 2, weight="bold")
        
        h2_font = main_font.copy()
        h2_font.configure(size=round(main_font.cget("size") * 1.5), weight="bold")
        
        h3_font = main_font.copy()
        h3_font.configure(size=main_font.cget("size"), weight="bold")
        
        self.text = tktextext.TweakableText(self, border=0, padx=15, pady=15,
                                            font=main_font,
                                            wrap="word")
        
        self.text.tag_configure("h1", font=h1_font)
        self.text.tag_configure("h2", font=h2_font)
        self.text.tag_configure("h3", font=h3_font)
        self.text.grid(row=0, column=0, sticky="nsew")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        
        self._vbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self._vbar.grid(row=0, column=1, sticky=tk.NSEW)
        self._vbar['command'] = self.text.yview 
        self.text['yscrollcommand'] = self._vbar.set  
        
        
        self.load_rst_file("help.rst")
        
    
    def clear(self):
        self.text.direct_delete("1.0", "end")
    
    def load_rst(self, source):
        
        def is_symbol_line(line, symbol, min_count=3):
            line = line.rstrip()
            return (line.startswith(symbol) 
                    and line.replace(symbol, "") == ""
                    and len(line) >= min_count)
        
        self.clear()
        lines = source.splitlines(True)
        i = 0
        while i < len(lines):
            if is_symbol_line(lines[i], "="):
                self.append_chars(lines[i+1], "h1")
                assert is_symbol_line(lines[i+2], "=")
                i += 3
            elif (i < len(lines)-1 
                  and is_symbol_line(lines[i+1], "=")):
                self.append_chars(lines[i], "h2")
                i += 2
            elif (i < len(lines)-1 
                  and is_symbol_line(lines[i+1], "-")):
                self.append_chars(lines[i], "h3")
                i += 2
            else:
                self.append_rst_line(lines[i])
                i += 1
    
    def append_chars(self, chars, tag=None):
        if tag:
            self.text.direct_insert("end", chars, (tag,))
        else:
            self.text.direct_insert("end", chars)
    
    def append_rst_line(self, source):
        self.append_chars(source)
    
    def load_rst_file(self, filename):
        if not os.path.isabs(filename):
            filename = os.path.join(os.path.dirname(__file__), filename) 
            
        with open(filename, encoding="UTF-8") as fp:
            self.load_rst(fp.read())


def open_help():
    get_workbench().show_view("HelpView")

def load_plugin():
    get_workbench().add_view(HelpView, "Help", "ne")
    get_workbench().add_command("help_contents", "help", "Help contents",
                                open_help,
                                group=30)
    


import os.path
import tkinter as tk
import time
from thonny.globals import get_workbench
from thonny.workbench import WorkbenchEvent
from datetime import datetime
import zipfile 
from tkinter.filedialog import asksaveasfilename
import json
from thonny.shell import ShellView
from thonny import THONNY_USER_DIR


class EventLogger:
    def __init__(self, filename):
        self._filename = filename
        self._init_logging()
        self._init_commands()
    
    def _init_commands(self):
        get_workbench().add_command(
            "export_usage_logs",
            "tools",
            "Export usage logs...",
            self._cmd_export,
            group=110
        )

    
    def _init_logging(self):
        self._events = []
        
        wb = get_workbench()
        wb.bind("WorkbenchClose", self._on_worbench_close, True)
        
        for sequence in ["<<Undo>>",
                         "<<Redo>>",
                         "<<Cut>>",
                         "<<Copy>>",
                         "<<Paste>>",
                         #"<<Selection>>",
                         #"<Key>",
                         #"<KeyRelease>",
                         "<Button-1>",
                         "<Button-2>",
                         "<Button-3>"
                         ]:
            self._bind_all(sequence)
        
        for sequence in ["Command",
                         "MagicCommand",
                         "Open",
                         "Save",
                         "SaveAs",
                         "NewFile",
                         "EditorTextCreated",
                         #"ShellTextCreated", # Too bad, this event happens before event_logging is loaded
                         "ShellCommand",
                         "ShellInput",
                         "ShowView",
                         "HideView",
                         "TextInsert",
                         "TextDelete",
                         ]:
            self._bind_workbench(sequence)

        self._bind_workbench("<FocusIn>", True)
        self._bind_workbench("<FocusOut>", True)
        
        ### log_user_event(KeyPressEvent(self, e.char, e.keysym, self.text.index(tk.INSERT)))

        
        # TODO: if event data includes an Editor, then look up also text id
    
    def _bind_workbench(self, sequence, only_workbench_widget=False):
        def handle(event):
            if not only_workbench_widget or event.widget == get_workbench():
                self._log_event(sequence, event)
        
        get_workbench().bind(sequence, handle, True)
    
    def _bind_all(self, sequence):
        
        def handle(event):
            self._log_event(sequence, event)
        
        tk._default_root.bind_all(sequence, handle, True)
    
    
    def _extract_interesting_data(self, event, sequence):
        attributes = vars(event)
        
        # generate some new attributes
        if "text_widget" not in attributes:
            if "editor" in attributes:
                attributes["text_widget"] = attributes["editor"].get_text_widget()
        
            if "widget" in attributes and isinstance(attributes["widget"], tk.Text):
                attributes["text_widget"] = attributes["widget"]
            
        
        if "text_widget" in attributes:
            widget = attributes["text_widget"]
            if isinstance(widget.master, ShellView):
                attributes["text_widget_context"] = "shell"
            
        
        # select attributes      
        data = {}
        for name in attributes:
            # skip some attributes
            if (name.startswith("_")
                or isinstance(event, WorkbenchEvent) and name in ["update", 
                                                                  "setdefault"]
                or isinstance(event, tk.Event) and name not in ["widget", 
                                                                "text_widget", 
                                                                "text_widget_context"]):
                continue
            
            value = attributes[name]
            
            if isinstance(value, tk.BaseWidget) or isinstance(value, tk.Tk):
                data[name + "_id"] = id(value)
                data[name + "_class"] = value.__class__.__name__
                
            elif type(value) in [str, int, float]:
                data[name] = value
            
            else:
                data[name] = repr(value)
        
        return data
    
    def _get_log_dir(self):
        return os.path.dirname(self._filename)
    
    def _cmd_export(self):
        
        filename = asksaveasfilename (
                filetypes =  [('Zip-files', '.zip'), ('all files', '.*')], 
                defaultextension = ".zip",
                initialdir = get_workbench().get_option("run.working_directory"),
                initialfile = time.strftime("ThonnyUsageLogs_%Y-%m-%d.zip")
        )
        
        if not filename:
            return
        
        log_dir = self._get_log_dir()
        
        with zipfile.ZipFile(filename, 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
            for item in os.listdir(log_dir):
                if item.endswith(".txt") or item.endswith(".zip"):
                    zipf.write(os.path.join(log_dir, item), arcname=item)
            
    
    def _log_event(self, sequence, event):
        data = self._extract_interesting_data(event, sequence)
        data["sequence"] = sequence 
        data["time"] = datetime.now().isoformat()
        self._events.append(data)
    
    def _on_worbench_close(self, event=None):
        with open(self._filename, encoding="UTF-8", mode="w") as fp:
            json.dump(self._events, fp, indent="    ")
        
        self._check_compress_logs()
    
    def _check_compress_logs(self):
        # if uncompressed logs have grown over 10MB,
        # compress these into new zipfile
        
        log_dir = self._get_log_dir()
        total_size = 0
        uncompressed_files = []
        for item in os.listdir(log_dir):
            if item.endswith(".txt"):
                full_name = os.path.join(log_dir, item)
                total_size += os.stat(full_name).st_size
                uncompressed_files.append((item, full_name))
            
        if total_size > 10*1024*1024:
            zip_filename = _generate_timestamp_file_name("zip")            
            with zipfile.ZipFile(zip_filename, 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
                for item, full_name in uncompressed_files:
                    zipf.write(full_name, arcname=item)
        
            for _, full_name in uncompressed_files:
                os.remove(full_name)
            

def _generate_timestamp_file_name(extension):
    # generate log filename
    folder = os.path.expanduser(os.path.join(THONNY_USER_DIR, "user_logs"))
    if not os.path.exists(folder):
        os.makedirs(folder)
        
    for i in range(100): 
        filename = os.path.join(folder, time.strftime("%Y-%m-%d_%H-%M-%S_{}.{}".format(i, extension)));
        if not os.path.exists(filename):
            return filename
    
    raise RuntimeError()

def load_plugin():
    filename = _generate_timestamp_file_name("txt")
    # create logger
    EventLogger(filename)
    


import tkinter as tk
from tkinter import font as tk_font 
from tkinter import ttk

from thonny.config_ui import ConfigurationPage
from thonny.globals import get_workbench
from thonny.ui_utils import create_string_var
import textwrap


class FontConfigurationPage(ConfigurationPage):
    
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        self._family_variable = create_string_var(
            get_workbench().get_option("view.editor_font_family"),
            modification_listener=self._update_preview_font)
        
        self._size_variable = create_string_var(
            get_workbench().get_option("view.editor_font_size"),
            modification_listener=self._update_preview_font)
        
        ttk.Label(self, text="Editor font").grid(row=0, column=0, sticky="w")
        
        self._family_combo = ttk.Combobox(self,
                                          exportselection=False,
                                          state='readonly',
                                          textvariable=self._family_variable,
                                          values=self._get_families_to_show())
        self._family_combo.grid(row=1, column=0, sticky=tk.NSEW, padx=(0,10))
        
        ttk.Label(self, text="Size").grid(row=0, column=1, sticky="w")
        self._size_combo = ttk.Combobox(self, width=4,
                                        exportselection=False,
                                        textvariable=self._size_variable,
                                        state='readonly',
                                        values=[str(x) for x in range(3,73)])
        
        self._size_combo.grid(row=1, column=1)
        
        
        ttk.Label(self, text="Preview").grid(row=2, column=0, sticky="w", pady=(10,0))
        self._preview_font = tk_font.Font()
        self._preview_text = tk.Text(self,
                                height=10,
                                borderwidth=1,
                                font=self._preview_font,
                                wrap=tk.WORD)
        self._preview_text.insert("1.0", textwrap.dedent("""
            The quick brown fox jumps over the lazy dog
            
            ABCDEFGHIJKLMNOPQRSTUVWXYZ
            abcdefghijklmnopqrstuvwxyz
            
            1234567890
            @$%()[]{}/\_-+
            "Hello " + 'world!'""").strip())
        self._preview_text.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW, pady=(0,5))
        
            
        self.columnconfigure(0, weight=1)
        self.rowconfigure(3, weight=1)
        self._update_preview_font()
    
    def apply(self):
        if (not self._family_variable.modified
            and not self._size_variable.modified):
            return
        
        get_workbench().set_option("view.editor_font_size", int(self._size_variable.get()))
        get_workbench().set_option("view.editor_font_family", self._family_variable.get())
        get_workbench().update_fonts()
    
    def _update_preview_font(self):
        self._preview_font.configure(family=self._family_variable.get(),
                                     size=int(self._size_variable.get()))
    
    def _get_families_to_show(self):
        return sorted(filter(
            lambda name : name[0].isalpha(),          
            tk_font.families()
        ))

def load_plugin():
    get_workbench().add_configuration_page("Font", FontConfigurationPage)


# -*- coding: utf-8 -*-

from tkinter import ttk, messagebox
from tkinter.dialog import Dialog

from thonny import tktextext, misc_utils
from thonny.globals import get_workbench
from thonny.misc_utils import running_on_mac_os, running_on_windows, running_on_linux
import tkinter as tk
import tkinter.messagebox as tkMessageBox
import traceback

import textwrap
import re
import collections
import threading
import signal
import subprocess
import os


CALM_WHITE = '#fdfdfd'

_images = set() # for keeping references to tkinter images to avoid garbace colleting them

class AutomaticPanedWindow(tk.PanedWindow):
    """
    Enables inserting panes according to their position_key-s.
    Automatically adds/removes itself to/from its master AutomaticPanedWindow.
    Fixes some style glitches.
    """ 
    def __init__(self, master, position_key=None,
                first_pane_size=1/3, last_pane_size=1/3, **kwargs):
        if not "sashwidth" in kwargs:
            kwargs["sashwidth"]=10
        
        if not "background" in kwargs:
            kwargs["background"] = get_main_background()
        
        tk.PanedWindow.__init__(self, master, **kwargs)
        
        self.position_key = position_key
        self.visible_panes = set()
        self.first_pane_size = first_pane_size
        self.last_pane_size = last_pane_size
        self._restoring_pane_sizes = False
        
        self._last_window_size = (0,0)
        self._full_size_not_final = True
        self._configure_binding = self.winfo_toplevel().bind("<Configure>", self._on_window_resize, True)
        self.bind("<B1-Motion>", self._on_mouse_dragged, True)
    
    def insert(self, pos, child, **kw):
        if pos == "auto":
            # According to documentation I should use self.panes()
            # but this doesn't return expected widgets
            for sibling in sorted(self.visible_panes, 
                                  key=lambda p:p.position_key 
                                        if hasattr(p, "position_key")
                                        else 0):
                if (not hasattr(sibling, "position_key") 
                    or sibling.position_key == None
                    or sibling.position_key > child.position_key):
                    pos = sibling
                    break
            else:
                pos = "end"
            
        if isinstance(pos, tk.Widget):
            kw["before"] = pos
        self.add(child, **kw)

    def add(self, child, **kw):
        if not "minsize" in kw:
            kw["minsize"]=60
            
        tk.PanedWindow.add(self, child, **kw)
        self.visible_panes.add(child)
        self._update_visibility()
        self._check_restore_pane_sizes()
    
    def remove(self, child):
        tk.PanedWindow.remove(self, child)
        self.visible_panes.remove(child)
        self._update_visibility()
        self._check_restore_pane_sizes()
    
    def forget(self, child):
        tk.PanedWindow.forget(self, child)
        self.visible_panes.remove(child)
        self._update_visibility()
        self._check_restore_pane_sizes()
    
    def destroy(self):
        self.winfo_toplevel().unbind("<Configure>", self._configure_binding)
        tk.PanedWindow.destroy(self)
        
    
    def is_visible(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return self.winfo_ismapped()
        else:
            return self in self.master.visible_panes
    
    def _on_window_resize(self, event):
        window = self.winfo_toplevel()
        window_size = (window.winfo_width(), window.winfo_height())
        initializing = hasattr(window, "initializing") and window.initializing
        
        if (not initializing
            and not self._restoring_pane_sizes 
            and (window_size != self._last_window_size or self._full_size_not_final)):
            self._check_restore_pane_sizes()
            self._last_window_size = window_size
    
    def _on_mouse_dragged(self, event):
        if event.widget == self and not self._restoring_pane_sizes:
            self._store_pane_sizes()
            
    
    def _store_pane_sizes(self):
        if len(self.panes()) > 1:
            self.last_pane_size = self._get_pane_size("last")
            if len(self.panes()) > 2:
                self.first_pane_size = self._get_pane_size("first")
    
    def _check_restore_pane_sizes(self):
        """last (and maybe first) pane sizes are stored, first (or middle)
        pane changes its size when window is resized"""
        
        window = self.winfo_toplevel()
        if hasattr(window, "initializing") and window.initializing:
            return
        
        try:
            self._restoring_pane_sizes = True
            if len(self.panes()) > 1:
                self._set_pane_size("last", self.last_pane_size)
                if len(self.panes()) > 2:
                    self._set_pane_size("first", self.first_pane_size)
        finally:
            self._restoring_pane_sizes = False
    
    def _get_pane_size(self, which):
        self.update_idletasks()
        
        if which == "first":
            coord = self.sash_coord(0)
        else:
            coord = self.sash_coord(len(self.panes())-2)
            
        if self.cget("orient") == tk.HORIZONTAL:
            full_size = self.winfo_width()
            sash_distance = coord[0]
        else:
            full_size = self.winfo_height()
            sash_distance = coord[1]
        
        if which == "first":
            return sash_distance
        else:
            return full_size - sash_distance 
        
    
    def _set_pane_size(self, which, size):
        #print("setsize", which, size)
        self.update_idletasks()
        
        if self.cget("orient") == tk.HORIZONTAL:
            full_size = self.winfo_width()
        else:
            full_size = self.winfo_height()
        
        self._full_size_not_final = full_size == 1
        
        if self._full_size_not_final:
            return
        
        if isinstance(size, float):
            size = int(full_size * size)
        
        #print("full vs size", full_size, size)
        
        if which == "first":
            sash_index = 0
            sash_distance = size 
        else:
            sash_index = len(self.panes())-2
            sash_distance = full_size - size 
        
        if self.cget("orient") == tk.HORIZONTAL:
            self.sash_place(sash_index, sash_distance, 0)
            #print("PLACE", sash_index, sash_distance, 0)
        else:
            self.sash_place(sash_index, 0, sash_distance)
            #print("PLACE", sash_index, 0, sash_distance)
      
    
    def _update_visibility(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return
        
        if len(self.visible_panes) == 0 and self.is_visible():
            self.master.forget(self)
            
        if len(self.panes()) > 0 and not self.is_visible():
            self.master.insert("auto", self)
        
    

class AutomaticNotebook(ttk.Notebook):
    """
    Enables inserting views according to their position keys.
    Remember its own position key. Automatically updates its visibility.
    """
    def __init__(self, master, position_key):
        ttk.Notebook.__init__(self, master)
        self.position_key = position_key
    
    def add(self, child, **kw):
        ttk.Notebook.add(self, child, **kw)
        self._update_visibility()
    
    def insert(self, pos, child, **kw):
        if pos == "auto":
            for sibling in map(self.nametowidget, self.tabs()):
                if (not hasattr(sibling, "position_key") 
                    or sibling.position_key == None
                    or sibling.position_key > child.position_key):
                    pos = sibling
                    break
            else:
                pos = "end"
            
        ttk.Notebook.insert(self, pos, child, **kw)
        self._update_visibility()
    
    def hide(self, tab_id):
        ttk.Notebook.hide(self, tab_id)
        self._update_visibility()
    
    def forget(self, tab_id):
        ttk.Notebook.forget(self, tab_id)
        self._update_visibility()
    
    def is_visible(self):
        return self in self.master.visible_panes
    
    def get_visible_child(self):
        for child in self.winfo_children():
            if str(child) == str(self.select()):
                return child
            
        return None
        
        
    def _update_visibility(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return
        if len(self.tabs()) == 0 and self.is_visible():
            self.master.remove(self)
            
        if len(self.tabs()) > 0 and not self.is_visible():
            self.master.insert("auto", self)
        

class TreeFrame(ttk.Frame):
    def __init__(self, master, columns, displaycolumns='#all', show_scrollbar=True):
        ttk.Frame.__init__(self, master)
        self.vert_scrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        if show_scrollbar:
            self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW)
        
        self.tree = ttk.Treeview(self, columns=columns, displaycolumns=displaycolumns, 
                                 yscrollcommand=self.vert_scrollbar.set)
        self.tree['show'] = 'headings'
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.vert_scrollbar['command'] = self.tree.yview
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.tree.bind("<<TreeviewSelect>>", self.on_select, "+")
        self.tree.bind("<Double-Button-1>", self.on_double_click, "+")
        
    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)
    
    def on_select(self, event):
        pass
    
    def on_double_click(self, event):
        pass



def sequence_to_accelerator(sequence):
    """Translates Tk event sequence to customary shortcut string
    for showing in the menu"""
    
    if not sequence:
        return ""
    
    if not sequence.startswith("<"):
        return sequence
    
    accelerator = (sequence
        .strip("<>")
        .replace("Key-", "")
        .replace("KeyPress-", "")
        .replace("Control", "Ctrl")
    )
    
    # Tweaking individual parts
    parts = accelerator.split("-")
    # tkinter shows shift with capital letter, but in shortcuts it's customary to include it explicitly
    if len(parts[-1]) == 1 and parts[-1].isupper() and not "Shift" in parts:
        parts.insert(-1, "Shift")
    
    # even when shift is not required, it's customary to show shortcut with capital letter
    if len(parts[-1]) == 1:
        parts[-1] = parts[-1].upper()
    
    accelerator = "+".join(parts)
    
    # Post processing
    accelerator = (accelerator
        .replace("Minus", "-").replace("minus", "-")
        .replace("Plus", "+").replace("plus", "+"))
    
    return accelerator
    

        
def get_zoomed(toplevel):
    if "-zoomed" in toplevel.wm_attributes(): # Linux
        return bool(toplevel.wm_attributes("-zoomed"))
    else: # Win/Mac
        return toplevel.wm_state() == "zoomed"
          

def set_zoomed(toplevel, value):
    if "-zoomed" in toplevel.wm_attributes(): # Linux
        toplevel.wm_attributes("-zoomed", str(int(value)))
    else: # Win/Mac
        if value:
            toplevel.wm_state("zoomed")
        else:
            toplevel.wm_state("normal")

class EnhancedTextWithLogging(tktextext.EnhancedText):
    def direct_insert(self, index, chars, tags=()):
        try:
            # try removing line numbers
            # TODO: shouldn't it take place only on paste?
            # TODO: does it occur when opening a file with line numbers in it?
            #if self._propose_remove_line_numbers and isinstance(chars, str):
            #    chars = try_remove_linenumbers(chars, self)
            concrete_index = self.index(index)
            return tktextext.EnhancedText.direct_insert(self, index, chars, tags=tags)
        finally:
            get_workbench().event_generate("TextInsert", index=concrete_index, 
                                           text=chars, tags=tags, text_widget=self)

    
    def direct_delete(self, index1, index2=None):
        try:
            # index1 may be eg "sel.first" and it doesn't make sense *after* deletion
            concrete_index1 = self.index(index1)
            if index2 is not None:
                concrete_index2 = self.index(index2)
            else:
                concrete_index2 = None
                
            return tktextext.EnhancedText.direct_delete(self, index1, index2=index2)
        finally:
            get_workbench().event_generate("TextDelete", index1=concrete_index1,
                                           index2=concrete_index2, text_widget=self)
            
    
class SafeScrollbar(ttk.Scrollbar):
    def set(self, first, last):
        try:
            ttk.Scrollbar.set(self, first, last)
        except:
            traceback.print_exc()

class AutoScrollbar(SafeScrollbar):
    # http://effbot.org/zone/tkinter-autoscrollbar.htm
    # a vert_scrollbar that hides itself if it's not needed.  only
    # works if you use the grid geometry manager.
    def set(self, lo, hi):
        # TODO: this can make GUI hang or max out CPU when scrollbar wobbles back and forth
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
        ttk.Scrollbar.set(self, lo, hi)
    def pack(self, **kw):
        raise tk.TclError("cannot use pack with this widget")
    def place(self, **kw):
        raise tk.TclError("cannot use place with this widget")

def update_entry_text(entry, text):
    original_state = entry.cget("state")
    entry.config(state="normal")
    entry.delete(0, "end")
    entry.insert(0, text)
    entry.config(state=original_state)


class ScrollableFrame(tk.Frame):
    # http://tkinter.unpythonic.net/wiki/VerticalScrolledFrame
    
    def __init__(self, master):
        tk.Frame.__init__(self, master, bg=CALM_WHITE)
        
        # set up scrolling with canvas
        vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.canvas = tk.Canvas(self, bg=CALM_WHITE, bd=0, highlightthickness=0,
                           yscrollcommand=vscrollbar.set)
        vscrollbar.config(command=self.canvas.yview)
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        vscrollbar.grid(row=0, column=1, sticky=tk.NSEW)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        
        self.interior = tk.Frame(self.canvas, bg=CALM_WHITE)
        self.interior.columnconfigure(0, weight=1)
        self.interior.rowconfigure(0, weight=1)
        self.interior_id = self.canvas.create_window(0,0, 
                                                    window=self.interior, 
                                                    anchor=tk.NW)
        self.bind('<Configure>', self._configure_interior, "+")
        self.bind('<Expose>', self._expose, "+")
        
    def _expose(self, event):
        self.update_idletasks()
        self._configure_interior(event)
    
    def _configure_interior(self, event):
        # update the scrollbars to match the size of the inner frame
        size = (self.canvas.winfo_width() , self.interior.winfo_reqheight())
        self.canvas.config(scrollregion="0 0 %s %s" % size)
        if (self.interior.winfo_reqwidth() != self.canvas.winfo_width()
            and self.canvas.winfo_width() > 10):
            # update the interior's width to fit canvas
            #print("CAWI", self.canvas.winfo_width())
            self.canvas.itemconfigure(self.interior_id,
                                      width=self.canvas.winfo_width())

class TtkDialog(Dialog):
    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = ttk.Frame(self)

        w = ttk.Button(box, text="OK", width=10, command=self.ok, default=tk.ACTIVE)
        w.pack(side=tk.LEFT, padx=5, pady=5)
        w = ttk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok, True)
        self.bind("<Escape>", self.cancel, True)

        box.pack()

    

class _QueryDialog(TtkDialog):

    def __init__(self, title, prompt,
                 initialvalue=None,
                 minvalue = None, maxvalue = None,
                 master = None,
                 selection_range=None):

        if not master:
            master = tk._default_root

        self.prompt   = prompt
        self.minvalue = minvalue
        self.maxvalue = maxvalue

        self.initialvalue = initialvalue
        self.selection_range = selection_range

        Dialog.__init__(self, master, title)

    def destroy(self):
        self.entry = None
        Dialog.destroy(self)

    def body(self, master):

        w = ttk.Label(master, text=self.prompt, justify=tk.LEFT)
        w.grid(row=0, padx=5, sticky=tk.W)

        self.entry = ttk.Entry(master, name="entry")
        self.entry.grid(row=1, padx=5, sticky="we")

        if self.initialvalue is not None:
            self.entry.insert(0, self.initialvalue)
            
            if self.selection_range:
                self.entry.icursor(self.selection_range[0])
                self.entry.select_range(self.selection_range[0], self.selection_range[1])
            else:
                self.entry.select_range(0, tk.END)

        return self.entry

    def validate(self):
        try:
            result = self.getresult()
        except ValueError:
            messagebox.showwarning(
                "Illegal value",
                self.errormessage + "\nPlease try again",
                parent = self
            )
            return 0

        if self.minvalue is not None and result < self.minvalue:
            messagebox.showwarning(
                "Too small",
                "The allowed minimum value is %s. "
                "Please try again." % self.minvalue,
                parent = self
            )
            return 0

        if self.maxvalue is not None and result > self.maxvalue:
            messagebox.showwarning(
                "Too large",
                "The allowed maximum value is %s. "
                "Please try again." % self.maxvalue,
                parent = self
            )
            return 0

        self.result = result

        return 1

class _QueryString(_QueryDialog):
    def __init__(self, *args, **kw):
        if "show" in kw:
            self.__show = kw["show"]
            del kw["show"]
        else:
            self.__show = None
        _QueryDialog.__init__(self, *args, **kw)

    def body(self, master):
        entry = _QueryDialog.body(self, master)
        if self.__show is not None:
            entry.configure(show=self.__show)
        return entry

    def getresult(self):
        return self.entry.get()


class ToolTip(object):
    """Taken from http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml"""

    def __init__(self, widget, options):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0
        self.options = options

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, _, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 27
        y = y + cy + self.widget.winfo_rooty() +27
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        if running_on_mac_os():
            # TODO: maybe it's because of Tk 8.5, not because of Mac
            tw.wm_transient(self.widget)
        tw.wm_geometry("+%d+%d" % (x, y))
        try:
            # For Mac OS
            tw.tk.call("::tk::unsupported::MacWindowStyle",
                       "style", tw._w,
                       "help", "noActivates")
        except tk.TclError:
            pass        
        label = tk.Label(tw, text=self.text, **self.options)
        label.pack()

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()

def create_tooltip(widget, text,
                   background="#ffffe0", relief=tk.SOLID, borderwidth=1, padx=1, pady=0,
                   **kw):
    options = kw.copy()
    options["background"] = background
    options["relief"] = relief
    options["borderwidth"] = borderwidth
    options["padx"] = padx
    options["pady"] = pady
    
    toolTip = ToolTip(widget, options)
    def enter(event):
        toolTip.showtip(text)
    def leave(event):
        toolTip.hidetip()
    widget.bind('<Enter>', enter)
    widget.bind('<Leave>', leave)

def askstring(title, prompt, **kw):
    '''get a string from the user

    Arguments:

        title -- the dialog title
        prompt -- the label text
        **kw -- see SimpleDialog class

    Return value is a string
    '''
    d = _QueryString(title, prompt, **kw)
    return d.result


def get_current_notebook_tab_widget(notebook):    
    for child in notebook.winfo_children():
        if str(child) == str(notebook.select()):
            return child
        
    return None

def create_string_var(value, modification_listener=None):
    """Creates a tk.StringVar with "modified" attribute
    showing whether the variable has been modified after creation"""
    return _create_var(tk.StringVar, value, modification_listener)

def create_int_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.IntVar, value, modification_listener)

def create_double_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.DoubleVar, value, modification_listener)

def create_boolean_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.BooleanVar, value, modification_listener)

def _create_var(class_, value, modification_listener):
    var = class_(value=value)
    var.modified = False
    
    def on_write(*args):
        var.modified = True
        if modification_listener:
            try:
                modification_listener()
            except:
                # Otherwise whole process will be brought down
                # because for some reason Tk tries to call non-existing method
                # on variable
                get_workbench().report_exception()
    
    # TODO: https://bugs.python.org/issue22115 (deprecation warning)
    var.trace("w", on_write)
    return var

def shift_is_pressed(event_state):
    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
    # http://stackoverflow.com/q/32426250/261181
    return event_state & 0x0001

def control_is_pressed(event_state):
    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
    # http://stackoverflow.com/q/32426250/261181
    return event_state & 0x0004

def select_sequence(win_version, mac_version, linux_version=None):
    if running_on_windows():
        return win_version
    elif running_on_mac_os():
        return mac_version
    elif running_on_linux() and linux_version:
        return linux_version
    else:
        return win_version

def try_remove_linenumbers(text, master):
    try:        
        if has_line_numbers(text) and tkMessageBox.askyesno (
                  title="Remove linenumbers",
                  message="Do you want to remove linenumbers from pasted text?",
                  default=tkMessageBox.YES,
                  master=master):
            return remove_line_numbers(text)
        else:
            return text
    except:
        traceback.print_exc()
        return text


def has_line_numbers(text):
    lines = text.splitlines()
    return (len(lines) > 2 
            and all([len(split_after_line_number(line)) == 2 for line in lines]))

def split_after_line_number(s): 
    parts = re.split("(^\s*\d+\.?)", s)
    if len(parts) == 1:
        return parts
    else:
        assert len(parts) == 3 and parts[0] == ''
        return parts[1:]

def remove_line_numbers(s):
    cleaned_lines = []
    for line in s.splitlines():
        parts = split_after_line_number(line)
        if len(parts) != 2:
            return s
        else:
            cleaned_lines.append(parts[1])
    
    return textwrap.dedent(("\n".join(cleaned_lines)) + "\n")
    
def get_main_background():
    main_background_option = get_workbench().get_option("theme.main_background")
    if main_background_option is not None:
        return main_background_option
    else:    
        theme = ttk.Style().theme_use()
        
        if theme == "clam":
            return "#dcdad5"
        elif theme == "aqua":
            return "systemSheetBackground"
        else: 
            return "SystemButtonFace"
    
def get_dialog_background_color():    
    theme = ttk.Style().theme_use()
    
    if theme == "aqua":
        return "systemSheetBackground"
    else: 
        return "SystemButtonFace"

def center_window(win, master=None):
    # looks like it doesn't take window border into account
    win.update_idletasks()
    
    if getattr(master, "initializing", False):
        # can't get reliable positions when main window is not in mainloop yet
        left = (win.winfo_screenwidth() - 600) // 2
        top = (win.winfo_screenheight() - 400) // 2
    else:
        if master is None:
            left = win.winfo_screenwidth() - win.winfo_width() // 2
            top = win.winfo_screenheight() - win.winfo_height() // 2
        else:
            left = master.winfo_rootx() + master.winfo_width() // 2 - win.winfo_width() // 2
            top = master.winfo_rooty() + master.winfo_height() // 2 - win.winfo_height() // 2
        
    win.geometry("+%d+%d" % (left, top))

class BusyTk(tk.Tk):
    def __init__(self, async_result, description, title="Please wait!"):
        self._async_result = async_result
        tk.Tk.__init__(self)
        self.update_idletasks()
        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()
        win_width = screen_width // 3
        win_height = screen_height // 3
        x = screen_width//2 - win_width//2
        y = screen_height//2 - win_height//2
        self.geometry("%dx%d+%d+%d" % (win_width, win_height, x, y))        
        
        main_frame = ttk.Frame(self)
        main_frame.grid(sticky=tk.NSEW, ipadx=15, ipady=15)
        main_frame.rowconfigure(0, weight=1)
        main_frame.columnconfigure(0, weight=1)
        self.title(title)
        self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.protocol("WM_DELETE_WINDOW", self._ok)
        self.desc_label = ttk.Label(main_frame, text=description)
        self.desc_label.grid(padx=20, pady=20, sticky="nsew")
        
        self.update_idletasks()
        self.after(500, self._poll)
    
    def _poll(self):
        if self._async_result.ready():
            self._ok()
        else:
            self.after(500, self._poll)
            self.desc_label["text"] = self.desc_label["text"] + "."
    
    def _ok(self):
        self.destroy() 


def run_with_busy_window(action, args=(), description=""):
    # http://stackoverflow.com/a/14299004/261181
    from multiprocessing.pool import ThreadPool
    pool = ThreadPool(processes=1)
    
    async_result = pool.apply_async(action, args) 
    dlg = BusyTk(async_result, description=description)
    dlg.mainloop()
    
    return async_result.get()  


class SubprocessDialog(tk.Toplevel):
    """Shows incrementally the output of given subprocess.
    Allows cancelling"""
    
    def __init__(self, master, proc, title, long_description=None, autoclose=True):
        self._proc = proc
        self.stdout = ""
        self.stderr = ""
        self._stdout_thread = None
        self._stderr_thread = None
        self.returncode = None
        self.cancelled = False
        self._autoclose = autoclose
        self._event_queue = collections.deque()
        
        tk.Toplevel.__init__(self, master)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        main_frame = ttk.Frame(self) # To get styled background
        main_frame.grid(sticky="nsew")

        text_font=tk.font.nametofont("TkFixedFont").copy()
        text_font["size"] = int(text_font["size"] * 0.9)
        text_font["family"] = "Courier" if running_on_mac_os() else "Courier New"
        text_frame = tktextext.TextFrame(main_frame, read_only=True, horizontal_scrollbar=False,
                                         background=get_main_background(),
                                         font=text_font,
                                         wrap="word")
        text_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=15, pady=15)
        self.text = text_frame.text
        self.text["width"] = 60
        self.text["height"] = 7
        if long_description is not None:
            self.text.direct_insert("1.0", long_description + "\n\n")
        
        self.button = ttk.Button(main_frame, text="Cancel", command=self._close)
        self.button.grid(row=1, column=0, pady=(0,15))
        
        main_frame.rowconfigure(0, weight=1)
        main_frame.columnconfigure(0, weight=1)
        

        self.title(title)
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        #self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.transient(master)
        self.grab_set() # to make it active and modal
        self.text.focus_set()
        
        
        self.bind('<Escape>', self._close_if_done, True) # escape-close only if process has completed 
        self.protocol("WM_DELETE_WINDOW", self._close)
        center_window(self, master)
        
        self._start_listening()
    
    def _start_listening(self):
       
        def listen_stream(stream_name):
            stream = getattr(self._proc, stream_name)
            while True:
                data = stream.readline()
                self._event_queue.append((stream_name, data))
                setattr(self, stream_name, getattr(self, stream_name) + data)
                if data == '':
                    break
            
            self.returncode = self._proc.wait()
        
        self._stdout_thread = threading.Thread(target=listen_stream, args=["stdout"])
        self._stdout_thread.start()
        if self._proc.stderr is not None:
            self._stderr_thread = threading.Thread(target=listen_stream, args=["stderr"])
            self._stderr_thread.start()
        
        def poll_output_events():
            while len(self._event_queue) > 0:
                stream_name, data = self._event_queue.popleft()
                self.text.direct_insert("end", data, tags=(stream_name, ))
                self.text.see("end")
            
            self.returncode = self._proc.poll() 
            if self.returncode == None:
                self.after(200, poll_output_events)
            else:
                self.button["text"] = "OK"
                self.button.focus_set()
                if self.returncode != 0:
                    self.text.direct_insert("end", "\n\nReturn code: ", ("stderr", ))
                elif self._autoclose:
                    self._close()
        
        poll_output_events()
        
    
    def _close_if_done(self, event):
        if self._proc.poll() is not None:
            self._close(event)        

    def _close(self, event=None):
        if self._proc.poll() is None:
            if messagebox.askyesno("Cancel the process?",
                "The process is still running.\nAre you sure you want to cancel?"):
                # try gently first
                try:
                    if running_on_windows():
                        os.kill(self._proc.pid, signal.CTRL_BREAK_EVENT)  # @UndefinedVariable
                    else:
                        os.kill(self._proc.pid, signal.SIGINT)
                        
                    self._proc.wait(2)
                except subprocess.TimeoutExpired:
                    if self._proc.poll() is None:
                        # now let's be more concrete
                        self._proc.kill()
                
                
                self.cancelled = True
                # Wait for threads to finish
                self._stdout_thread.join(2)
                if self._stderr_thread is not None:
                    self._stderr_thread.join(2)
                
                # fetch output about cancelling
                while len(self._event_queue) > 0:
                    stream_name, data = self._event_queue.popleft()
                    self.text.direct_insert("end", data, tags=(stream_name, ))
                self.text.direct_insert("end", "\n\nPROCESS CANCELLED")
                self.text.see("end")
                    
                    
            else:
                return
        else:
            self.destroy()

def get_busy_cursor():
    if running_on_windows():
        return "wait"
    elif running_on_mac_os():
        return "spinning"
    else:
        return "watch"

def get_tk_version_str():
    return tk._default_root.tk.call('info', 'patchlevel')

def get_tk_version_info():
    result = []
    for part in get_tk_version_str().split("."):
        try:
            result.append(int(part))
        except:
            result.append(0)
    return tuple(result) 



# -*- coding: utf-8 -*-

import os.path
import platform
import sys
import shutil
import time


def eqfn(name1, name2):
    return os.path.normcase(name1) == os.path.normcase(name2)

def delete_dir_try_hard(path, hardness=5):
    # Deleting the folder on Windows is not so easy task
    # http://bugs.python.org/issue15496
    for i in range(hardness):
        if os.path.exists(path):
            time.sleep(i * 0.5)
            shutil.rmtree(path, True)
        else:
            break

    if os.path.exists(path):
        # try once more but now without ignoring errors
        shutil.rmtree(path, False)

def running_on_windows():
    return platform.system() == "Windows"
    
def running_on_mac_os():
    return platform.system() == "Darwin"
    
def running_on_linux():
    return platform.system() == "Linux"

def is_hidden_or_system_file(path):
    if os.path.basename(path).startswith("."):
        return True
    elif running_on_windows():
        from ctypes import windll
        FILE_ATTRIBUTE_HIDDEN = 0x2
        FILE_ATTRIBUTE_SYSTEM = 0x4
        return bool(windll.kernel32.GetFileAttributesW(path)  # @UndefinedVariable
                & (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM))
    else:
        return False 
    
def get_win_drives():
    # http://stackoverflow.com/a/2288225/261181
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa364939%28v=vs.85%29.aspx
    import string
    from ctypes import windll
    
    all_drive_types = ['DRIVE_UNKNOWN', 
                       'DRIVE_NO_ROOT_DIR',
                       'DRIVE_REMOVABLE',
                       'DRIVE_FIXED',
                       'DRIVE_REMOTE',
                       'DRIVE_CDROM',
                       'DRIVE_RAMDISK']
    
    required_drive_types = ['DRIVE_REMOVABLE',
                            'DRIVE_FIXED',
                            'DRIVE_REMOTE',
                            'DRIVE_RAMDISK']

    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()  # @UndefinedVariable
    for letter in string.ascii_uppercase:
        drive_type = all_drive_types[windll.kernel32.GetDriveTypeW("%s:\\" % letter)]  # @UndefinedVariable
        if bitmask & 1 and drive_type in required_drive_types:
            drives.append(letter + ":\\")
        bitmask >>= 1

    return drives



def shorten_repr(original_repr, max_len=1000):
    if len(original_repr) > max_len:
        return original_repr[:max_len] + " ... [{} chars truncated]".format(len(original_repr) - max_len)
    else:
        return original_repr
        
def __maybe_later_get_thonny_data_folder():
    if running_on_windows():
        # CSIDL_LOCAL_APPDATA 
        # http://www.installmate.com/support/im9/using/symbols/functions/csidls.htm
        return os.path.join(__maybe_later_get_windows_special_folder(28), "Thonny")
    elif running_on_linux():
        # https://specifications.freedesktop.org/basedir-spec/latest/ar01s02.html
        # $XDG_DATA_HOME or $HOME/.local/share
        data_home = os.environ.get("XDG_DATA_HOME", 
                                   os.path.expanduser("~/.local/share"))
        return os.path.join(data_home, "Thonny") 
    elif running_on_mac_os():
        return os.path.expanduser("~/Library/Thonny")
    else:
        return os.path.expanduser("~/.thonny")

def __maybe_later_get_windows_special_folder(code):
    # http://stackoverflow.com/a/3859336/261181
    # http://www.installmate.com/support/im9/using/symbols/functions/csidls.htm
    import ctypes.wintypes
    SHGFP_TYPE_CURRENT= 0
    buf= ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
    ctypes.windll.shell32.SHGetFolderPathW(0, code, 0, SHGFP_TYPE_CURRENT, buf)
    return buf.value

def get_python_version_string(version_info=None):
    if version_info == None:
        version_info = sys.version_info
         
    result = ".".join(map(str, version_info[:3]))
    if version_info[3] != "final":
        result += "-" + version_info[3]
    
    result += " (" + ("64" if sys.maxsize > 2**32 else "32")+ " bit)\n"
    
    return result    




# -*- coding: utf-8 -*-

import os.path
import re
from tkinter import ttk
import traceback

import thonny
from thonny import memory, roughparse
from thonny.common import ToplevelCommand, parse_shell_command
from thonny.misc_utils import running_on_mac_os, shorten_repr
from thonny.ui_utils import EnhancedTextWithLogging
import tkinter as tk
from thonny.globals import get_workbench, get_runner
from thonny.codeview import EDIT_BACKGROUND, PythonText
from thonny.tktextext import index2line


class ShellView (ttk.Frame):
    def __init__(self, master, **kw):
        ttk.Frame.__init__(self, master, **kw)
        
        self.vert_scrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.vert_scrollbar.grid(row=0, column=2, sticky=tk.NSEW)
        self.text = ShellText(self,
                            font=get_workbench().get_font("EditorFont"),
                            #foreground="white",
                            #background="#666666",
                            highlightthickness=0,
                            #highlightcolor="LightBlue",
                            borderwidth=0,
                            yscrollcommand=self.vert_scrollbar.set,
                            padx=4,
                            insertwidth=2,
                            height=10,
                            undo=True)
        
        get_workbench().event_generate("ShellTextCreated", text_widget=self.text)
        get_workbench().add_command("clear_shell", "edit", "Clear shell",
                                    self.clear_shell,
                                    group=200)
        
        self.text.grid(row=0, column=1, sticky=tk.NSEW)
        self.vert_scrollbar['command'] = self.text.yview
        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)

    def focus_set(self):
        self.text.focus_set()
    
    def add_command(self, command, handler):
        self.text.add_command(command, handler)

    def submit_command(self, cmd_line):
        self.text.submit_command(cmd_line)
    
    def clear_shell(self):
        self.text._clear_shell()
        
    def report_exception(self, prelude=None, conclusion=None):
        if prelude is not None:
            self.text.direct_insert("end", prelude + "\n", ("stderr",))
        
        self.text.direct_insert("end", traceback.format_exc() + "\n", ("stderr",))
        
        if conclusion is not None:
            self.text.direct_insert("end", conclusion + "\n", ("stderr",))
        


class ShellText(EnhancedTextWithLogging, PythonText):
    
    def __init__(self, master, cnf={}, **kw):
        if not "background" in kw:
            kw["background"] = EDIT_BACKGROUND
            
        EnhancedTextWithLogging.__init__(self, master, cnf, **kw)
        self.bindtags(self.bindtags() + ('ShellText',))
        
        self._before_io = True
        self._command_history = [] # actually not really history, because each command occurs only once
        self._command_history_current_index = None
        
        
        """
        self.margin = tk.Text(self,
                width = 4,
                padx = 4,
                highlightthickness = 0,
                takefocus = 0,
                bd = 0,
                #font = self.font,
                cursor = "dotbox",
                background = '#e0e0e0',
                foreground = '#999999',
                #state='disabled'
                )
        self.margin.grid(row=0, column=0)
        """
        
        self.bind("<Up>", self._arrow_up, True)
        self.bind("<Down>", self._arrow_down, True)
        self.bind("<KeyPress>", self._text_key_press, True)
        self.bind("<KeyRelease>", self._text_key_release, True)
        
        prompt_font = get_workbench().get_font("BoldEditorFont")
        vert_spacing = 10
        io_indent = 16
        code_indent = prompt_font.measure(">>> ")
        
        
        self.tag_configure("toplevel", font=get_workbench().get_font("EditorFont"))
        self.tag_configure("prompt", foreground="purple", font=prompt_font)
        self.tag_configure("command", foreground="black",
                           lmargin1=code_indent, lmargin2=code_indent)
        self.tag_configure("welcome", foreground="DarkGray", font=get_workbench().get_font("EditorFont"))
        self.tag_configure("automagic", foreground="DarkGray", font=get_workbench().get_font("EditorFont"))
        self.tag_configure("value", foreground="DarkBlue") 
        self.tag_configure("error", foreground="Red")
        
        self.tag_configure("io", lmargin1=io_indent, lmargin2=io_indent, rmargin=io_indent,
                                font=get_workbench().get_font("IOFont"))
        self.tag_configure("stdin", foreground="Blue")
        self.tag_configure("stdout", foreground="Black")
        self.tag_configure("stderr", foreground="Red")
        self.tag_configure("hyperlink", foreground="#3A66DD", underline=True)
        self.tag_bind("hyperlink", "<ButtonRelease-1>", self._handle_hyperlink)
        self.tag_bind("hyperlink", "<Enter>", self._hyperlink_enter)
        self.tag_bind("hyperlink", "<Leave>", self._hyperlink_leave)
        
        self.tag_configure("vertically_spaced", spacing1=vert_spacing)
        self.tag_configure("inactive", foreground="#aaaaaa")
        
        # create 3 marks: input_start shows the place where user entered but not-yet-submitted
        # input starts, output_end shows the end of last output,
        # output_insert shows where next incoming program output should be inserted
        self.mark_set("input_start", "end-1c")
        self.mark_gravity("input_start", tk.LEFT)
        
        self.mark_set("output_end", "end-1c")
        self.mark_gravity("output_end", tk.LEFT)
        
        self.mark_set("output_insert", "end-1c")
        self.mark_gravity("output_insert", tk.RIGHT)
        
        
        self.active_object_tags = set()
        
        self._command_handlers = {}
        
        self._last_configuration = None
    
        get_workbench().bind("InputRequest", self._handle_input_request, True)
        get_workbench().bind("ProgramOutput", self._handle_program_output, True)
        get_workbench().bind("ToplevelResult", self._handle_toplevel_result, True)
        
        self._init_menu()
    
    def _init_menu(self):
        self._menu = tk.Menu(self, tearoff=False)
        self._menu.add_command(label="Clear shell", command=self._clear_shell)
    
    def add_command(self, command, handler):
        self._command_handlers[command] = handler
        
    def submit_command(self, cmd_line):
        assert get_runner().get_state() == "waiting_toplevel_command"
        self.delete("input_start", "end")
        self.insert("input_start", cmd_line, ("automagic",))
        self.see("end")
        self.mark_set("insert", "end")
        self._try_submit_input()
    
    def _handle_input_request(self, msg):
        self["font"] = get_workbench().get_font("IOFont") # otherwise the cursor is of toplevel size
        self.focus_set()
        self.mark_set("insert", "end")
        self.tag_remove("sel", "1.0", tk.END)
        self._current_input_request = msg
        self._try_submit_input() # try to use leftovers from previous request
        self.see("end")

    def _handle_program_output(self, msg):
        self["font"] = get_workbench().get_font("IOFont")
        
        # mark first line of io
        if self._before_io:
            self._insert_text_directly(msg.data[0], ("io", msg.stream_name, "vertically_spaced"))
            self._before_io = False
            self._insert_text_directly(msg.data[1:], ("io", msg.stream_name))
        else:
            self._insert_text_directly(msg.data, ("io", msg.stream_name))
        
        self.mark_set("output_end", self.index("end-1c"))
        self.see("end")
            
    def _handle_toplevel_result(self, msg):
        self["font"] = get_workbench().get_font("EditorFont")
        self._before_io = True
        if hasattr(msg, "error"):
            self._insert_text_directly(msg.error + "\n", ("toplevel", "error"))
        
        if hasattr(msg, "welcome_text"):
            configuration = get_workbench().get_option("run.backend_configuration") 
            welcome_text = msg.welcome_text
            if hasattr(msg, "executable") and msg.executable != thonny.running.get_private_venv_executable():
                welcome_text += " (" + msg.executable + ")"
            if (configuration != self._last_configuration
                and not (self._last_configuration is None and not configuration)):
                    self._insert_text_directly(welcome_text, ("welcome",))
                    
            self._last_configuration = get_workbench().get_option("run.backend_configuration")
            
        
        if hasattr(msg, "value_info"):
            value_repr = shorten_repr(msg.value_info["repr"], 10000)
            if value_repr != "None":
                if get_workbench().in_heap_mode():
                    value_repr = memory.format_object_id(msg.value_info["id"])
                object_tag = "object_" + str(msg.value_info["id"])
                self._insert_text_directly(value_repr + "\n", ("toplevel",
                                                               "value",
                                                               object_tag))
                if running_on_mac_os():
                    sequence = "<Command-Button-1>"
                else:
                    sequence = "<Control-Button-1>"
                self.tag_bind(object_tag, sequence,
                                   lambda _: get_workbench().event_generate(
                                        "ObjectSelect", object_id=msg.value_info["id"]))
                
                self.active_object_tags.add(object_tag)
        
        self.mark_set("output_end", self.index("end-1c"))
        self._insert_prompt()
        self._try_submit_input() # Trying to submit leftover code (eg. second magic command)
        self.see("end")
            
    def _insert_prompt(self):
        # if previous output didn't put a newline, then do it now
        if not self.index("output_insert").endswith(".0"):
            self._insert_text_directly("\n", ("io",))
        
        prompt_tags = ("toplevel", "prompt")
         
        # if previous line has value or io then add little space
        prev_line = self.index("output_insert - 1 lines")
        prev_line_tags = self.tag_names(prev_line)
        if "io" in prev_line_tags or "value" in prev_line_tags:
            prompt_tags += ("vertically_spaced",)
            #self.tag_add("last_result_line", prev_line)
        
        self._insert_text_directly(">>> ", prompt_tags)
        self.edit_reset();
    
    def intercept_insert(self, index, txt, tags=()):
        if (self._editing_allowed()
            and self._in_current_input_range(index)):
            #self._print_marks("before insert")
            # I want all marks to stay in place
            self.mark_gravity("input_start", tk.LEFT)
            self.mark_gravity("output_insert", tk.LEFT)
            
            if get_runner().get_state() == "waiting_input":
                tags = tags + ("io", "stdin")
            else:
                tags = tags + ("toplevel", "command")
            
            EnhancedTextWithLogging.intercept_insert(self, index, txt, tags)
            
            if get_runner().get_state() == "waiting_input":
                if self._before_io:
                    # tag first char of io differently
                    self.tag_add("vertically_spaced", index)
                    self._before_io = False
                    
                self._try_submit_input()
            
            self.see("insert")
        else:
            self.bell()
            
    def intercept_delete(self, index1, index2=None, **kw):
        if index1 == "sel.first" and index2 == "sel.last" and not self.has_selection():
            return
        
        if (self._editing_allowed() 
            and self._in_current_input_range(index1)
            and (index2 is None or self._in_current_input_range(index2))):
            self.direct_delete(index1, index2, **kw)
        else:
            self.bell()
    
    def perform_return(self, event):
        if get_runner().get_state() == "waiting_input":
            # if we are fixing the middle of the input string and pressing ENTER
            # then we expect the whole line to be submitted not linebreak to be inserted
            # (at least that's how IDLE works)
            self.mark_set("insert", "end") # move cursor to the end
            
            # Do the return without auto indent
            EnhancedTextWithLogging.perform_return(self, event)
             
            self._try_submit_input()
            
        elif get_runner().get_state() == "waiting_toplevel_command":
            # Same with editin middle of command, but only if it's a single line command
            whole_input = self.get("input_start", "end-1c") # asking the whole input
            if ("\n" not in whole_input
                and self._code_is_ready_for_submission(whole_input)):
                self.mark_set("insert", "end") # move cursor to the end
                # Do the return without auto indent
                EnhancedTextWithLogging.perform_return(self, event)
            else:
                # Don't want auto indent when code is ready for submission
                source = self.get("input_start", "insert")
                tail = self.get("insert", "end")
                
                if self._code_is_ready_for_submission(source + "\n", tail):
                    # No auto-indent
                    EnhancedTextWithLogging.perform_return(self, event)
                else:
                    # Allow auto-indent
                    PythonText.perform_return(self, event)
                
            self._try_submit_input()
            
        return "break"
    
    def on_secondary_click(self, event):
        super().on_secondary_click(event)
        self._menu.tk_popup(event.x_root, event.y_root)
        
    def _in_current_input_range(self, index):
        try:
            return self.compare(index, ">=", "input_start")
        except:
            return False
    
    def _insert_text_directly(self, txt, tags=()):
        def _insert(txt, tags):
            if txt != "":
                self.direct_insert("output_insert", txt, tags)
                
        # I want the insertion to go before marks 
        #self._print_marks("before output")
        self.mark_gravity("input_start", tk.RIGHT)
        self.mark_gravity("output_insert", tk.RIGHT)
        tags = tuple(tags)
        
        if "stderr" in tags or "error" in tags:
            # show lines pointing to source lines as hyperlinks
            for line in txt.splitlines(True):
                parts = re.split(r'(File .* line \d+.*)$', line, maxsplit=1)
                if len(parts) == 3 and "<pyshell" not in line:
                    _insert(parts[0], tags)
                    _insert(parts[1], tags + ("hyperlink",))
                    _insert(parts[2], tags)
                else:
                    _insert(line, tags)
        else:
            _insert(txt, tags)
            
        #self._print_marks("after output")
        # output_insert mark will move automatically because of its gravity
    
    
    def _try_submit_input(self):
        # see if there is already enough inputted text to submit
        input_text = self.get("input_start", "insert")
        tail = self.get("insert", "end")
        
        # user may have pasted more text than necessary for this request
        submittable_text = self._extract_submittable_input(input_text, tail)
        
        if submittable_text is not None:
            if get_runner().get_state() == "waiting_toplevel_command":
                # clean up the tail
                if len(tail) > 0:
                    assert tail.strip() == ""
                    self.delete("insert", "end-1c")
                    
            
            # leftover text will be kept in widget, waiting for next request.
            start_index = self.index("input_start")
            end_index = self.index("input_start+{0}c".format(len(submittable_text)))
            
            # apply correct tags (if it's leftover then it doesn't have them yet)
            if get_runner().get_state() == "waiting_input":
                self.tag_add("io", start_index, end_index)
                self.tag_add("stdin", start_index, end_index)
            else:
                self.tag_add("toplevel", start_index, end_index)
                self.tag_add("command", start_index, end_index)
                
            
            
            # update start mark for next input range
            self.mark_set("input_start", end_index)
            
            # Move output_insert mark after the requested_text
            # Leftover input, if any, will stay after output_insert, 
            # so that any output that will come in before
            # next input request will go before leftover text
            self.mark_set("output_insert", end_index)
            
            # remove tags from leftover text
            for tag in ("io", "stdin", "toplevel", "command"):
                # don't remove automagic, because otherwise I can't know it's auto 
                self.tag_remove(tag, end_index, "end")
                
            self._submit_input(submittable_text)
    
    def _editing_allowed(self):
        return get_runner().get_state() in ('waiting_toplevel_command', 'waiting_input')
    
    def _extract_submittable_input(self, input_text, tail):
        
        if get_runner().get_state() == "waiting_toplevel_command":
            if input_text.endswith("\n"):
                if input_text.strip().startswith("%"):
                    # if several magic command are submitted, then take only first
                    return input_text[:input_text.index("\n")+1]
                elif self._code_is_ready_for_submission(input_text, tail):
                    return input_text
                else:
                    return None
            else:
                return None
            
        elif get_runner().get_state() == "waiting_input":
            input_request = self._current_input_request
            method = input_request.method
            limit = input_request.limit
            # TODO: what about EOF?
            if isinstance(limit, int) and limit < 0:
                limit = None
            
            if method == "readline":
                # TODO: is it correct semantics?
                i = 0
                if limit == 0:
                    return ""
                
                while True:
                    if i >= len(input_text):
                        return None
                    elif limit is not None and i+1 == limit:
                        return input_text[:i+1]
                    elif input_text[i] == "\n":
                        return input_text[:i+1]
                    else:
                        i += 1
            else:
                raise AssertionError("only readline is supported at the moment")
    
    def _code_is_ready_for_submission(self, source, tail=""):
        # Ready to submit if ends with empty line 
        # or is complete single-line code
        
        if tail.strip() != "":
            return False
        
        # First check if it has unclosed parens, unclosed string or ending with : or \
        parser = roughparse.RoughParser(self.indentwidth, self.tabwidth)
        parser.set_str(source.rstrip() + "\n")
        if (parser.get_continuation_type() != roughparse.C_NONE
                or parser.is_block_opener()):
            return False
        
        # Multiline compound statements need to end with empty line to be considered
        # complete.
        lines = source.splitlines()
        # strip starting empty and comment lines
        while (len(lines) > 0
               and (lines[0].strip().startswith("#")
                    or lines[0].strip() == "")):
            lines.pop(0)
        
        compound_keywords = ["if", "while", "for", "with", "try", "def", "class", "async", "await"]
        if len(lines) > 0:
            first_word = lines[0].strip().split()[0]
            if (first_word in compound_keywords
                and not source.replace(" ", "").replace("\t", "").endswith("\n\n")):
                # last line is not empty
                return False
        
        return True
    
    def _submit_input(self, text_to_be_submitted):
        if get_runner().get_state() == "waiting_toplevel_command":
            # register in history and count
            if text_to_be_submitted in self._command_history:
                self._command_history.remove(text_to_be_submitted)
            self._command_history.append(text_to_be_submitted)
            self._command_history_current_index = None # meaning command selection is not in process
            
            try:
                if text_to_be_submitted.startswith("%"):
                    command, _ = parse_shell_command(text_to_be_submitted)
                    get_workbench().event_generate("MagicCommand", cmd_line=text_to_be_submitted)
                    if command in self._command_handlers:
                        self._command_handlers[command](text_to_be_submitted)
                        get_workbench().event_generate("AfterKnownMagicCommand", cmd_line=text_to_be_submitted)
                    else:
                        self._insert_text_directly("Unknown magic command: " + command)
                        self._insert_prompt()
                else:
                    get_runner().send_command(
                        ToplevelCommand(command="execute_source",
                                        source=text_to_be_submitted))
                
            except:
                get_workbench().report_exception()
                self._insert_prompt()
                
            get_workbench().event_generate("ShellCommand", command_text=text_to_be_submitted)
        else:
            assert get_runner().get_state() == "waiting_input"
            get_runner().send_program_input(text_to_be_submitted)
            get_workbench().event_generate("ShellInput", input_text=text_to_be_submitted)
    
    
    def _arrow_up(self, event):
        if not self._in_current_input_range("insert"):
            return

        insert_line = index2line(self.index("insert"))
        input_start_line = index2line(self.index("input_start"))
        if insert_line != input_start_line:
            # we're in the middle of a multiline command
            return
        
        if len(self._command_history) == 0 or self._command_history_current_index == 0:
            # can't take previous command
            return "break"
        
        if self._command_history_current_index is None:
            self._command_history_current_index = len(self._command_history)-1
        else:
            self._command_history_current_index -= 1
        
        cmd = self._command_history[self._command_history_current_index]
        if cmd[-1] == "\n": 
            cmd = cmd[:-1] # remove the submission linebreak
        self._propose_command(cmd)
        return "break"
    
    def _arrow_down(self, event):
        if not self._in_current_input_range("insert"):
            return
        
        insert_line = index2line(self.index("insert"))
        last_line = index2line(self.index("end-1c"))
        if insert_line != last_line:
            # we're in the middle of a multiline command
            return
        
        if (len(self._command_history) == 0 
            or self._command_history_current_index == len(self._command_history)-1):
            # can't take next command
            return "break"
        
        
        if self._command_history_current_index is None:
            self._command_history_current_index = len(self._command_history)-1
        else:
            self._command_history_current_index += 1

        self._propose_command(self._command_history[self._command_history_current_index].strip("\n"))
        return "break"
    
    def _propose_command(self, cmd_line):
        self.delete("input_start", "end")
        self.intercept_insert("input_start", cmd_line)
        self.see("insert")
    
    def _text_key_press(self, event):
        # TODO: this underline may confuse, when user is just copying on pasting
        # try to add this underline only when mouse is over the value
        """
        if event.keysym in ("Control_L", "Control_R", "Command"):  # TODO: check in Mac
            self.tag_configure("value", foreground="DarkBlue", underline=1)
        """
    
    def _text_key_release(self, event):
        if event.keysym in ("Control_L", "Control_R", "Command"):  # TODO: check in Mac
            self.tag_configure("value", foreground="DarkBlue", underline=0)

    def _clear_shell(self):
        end_index = self.index("output_end")
        self.direct_delete("1.0", end_index)

    def compute_smart_home_destination_index(self):
        """Is used by EnhancedText"""
        
        if self._in_current_input_range("insert"):
            # on input line, go to just after prompt
            return "input_start"
        else:
            return super().compute_smart_home_destination_index()
    
    def _hyperlink_enter(self, event):
        self.config(cursor="hand2")
        
    def _hyperlink_leave(self, event):
        self.config(cursor="")
        
    def _handle_hyperlink(self, event):
        try:
            line = self.get("insert linestart", "insert lineend")
            matches = re.findall(r'File "([^"]+)", line (\d+)', line)
            if len(matches) == 1 and len(matches[0]) == 2:
                filename, lineno = matches[0]
                lineno = int(lineno)
                if os.path.exists(filename) and os.path.isfile(filename):
                    # TODO: better use events instead direct referencing
                    get_workbench().get_editor_notebook().show_file(filename, lineno)
        except:
            traceback.print_exc()
    
    
    def _invalidate_current_data(self):
        """
        Grayes out input & output displayed so far
        """
        end_index = self.index("output_end")
        
        self.tag_add("inactive", "1.0", end_index)
        self.tag_remove("value", "1.0", end_index)
        
        while len(self.active_object_tags) > 0:
            self.tag_remove(self.active_object_tags.pop(), "1.0", "end")
        
        

    

    
    


import tkinter as tk
from thonny.globals import get_workbench
import logging
import thonny.jedi_utils as jedi_utils

class LocalsHighlighter:

    def __init__(self, text, local_variable_font=None):
        self.text = text
        
        if local_variable_font:
            self.local_variable_font=local_variable_font
        else:
            self.local_variable_font = self.text["font"]
        
        self._configure_tags()
        self._update_scheduled = False
    
    def get_positions_simple_but_incorrect(self):
        # goto_assignments only gives you last assignment to given node
        import jedi
        defs = jedi.names(self.text.get('1.0', 'end'), path="",
                           all_scopes=True, definitions=True, references=True)
        result = set()
        for definition in defs:
            if definition.parent().type == "function": # is located in a function
                ass = definition.goto_assignments()
                if len(ass) > 0 and ass[0].parent().type == "function": # is assigned to in a function
                    pos = ("%d.%d" % (definition.line, definition.column),
                           "%d.%d" % (definition.line, definition.column+len(definition.name)))
                    result.add(pos)
        return result
        
    
    def get_positions_correct_but_using_private_parts(self):
        from jedi import Script
        
        try:
            from jedi.parser.python import tree
        except ImportError:
            from jedi.parser import tree

        locs = []

        def process_scope(scope):
            if isinstance(scope, tree.Function):
                # process all children after name node,
                # (otherwise name of global function will be marked as local def)
                local_names = set() 
                global_names = set()
                for child in scope.children[2:]:
                    process_node(child, local_names, global_names)
            else:
                for child in scope.subscopes:
                    process_scope(child)
        
        def process_node(node, local_names, global_names):
            if isinstance(node, tree.GlobalStmt):
                global_names.update([n.value for n in node.get_global_names()])
                
            elif isinstance(node, tree.Name):
                if node.value in global_names:
                    return
                
                if node.is_definition(): # local def
                    locs.append(node)
                    local_names.add(node.value)
                elif node.value in local_names: # use of local 
                    locs.append(node)
                    
            elif isinstance(node, tree.BaseNode):
                # ref: jedi/parser/grammar*.txt
                if node.type == "trailer" and node.children[0].value == ".":
                    # this is attribute
                    return
                
                if isinstance(node, tree.Function):
                    global_names = set() # outer global statement doesn't have effect anymore
                
                for child in node.children:
                    process_node(child, local_names, global_names)

        source = self.text.get('1.0', 'end')
        script = Script(source + ")") # https://github.com/davidhalter/jedi/issues/897
        module = jedi_utils.get_module_node(script)
        for child in module.children:
            if isinstance(child, tree.BaseNode) and child.is_scope():
                process_scope(child)

        loc_pos = set(("%d.%d" % (usage.start_pos[0], usage.start_pos[1]),
                "%d.%d" % (usage.start_pos[0], usage.start_pos[1] + len(usage.value)))
                for usage in locs)

        return loc_pos

    def _configure_tags(self):
        self.text.tag_configure("LOCAL_NAME",
                                font=self.local_variable_font, 
                                foreground="#000055")
        self.text.tag_raise("sel")
        
    def _highlight(self, pos_info):
        for pos in pos_info:
            start_index, end_index = pos[0], pos[1]
            self.text.tag_add("LOCAL_NAME", start_index, end_index)

    def schedule_update(self):
        def perform_update():
            try:
                self.update()
            finally:
                self._update_scheduled = False
        
        if not self._update_scheduled:
            self._update_scheduled = True
            self.text.after_idle(perform_update)
            
    def update(self):
        self.text.tag_remove("LOCAL_NAME", "1.0", "end")
        
        if get_workbench().get_option("view.locals_highlighting"):
            try:
                highlight_positions = self.get_positions_correct_but_using_private_parts()
                self._highlight(highlight_positions)
            except:
                logging.exception("Problem when updating local variable tags")


def update_highlighting(event):
    assert isinstance(event.widget, tk.Text)
    text = event.widget
    
    if not hasattr(text, "local_highlighter"):
        text.local_highlighter = LocalsHighlighter(text,
            get_workbench().get_font("ItalicEditorFont"))
        
    text.local_highlighter.schedule_update()


def load_plugin():
    if jedi_utils.get_version_tuple() < (0, 9):
        logging.warning("Jedi version is too old. Disabling locals marker")
        return 
    
    wb = get_workbench()
    wb.set_default("view.locals_highlighting", True)
    wb.bind_class("CodeViewText", "<<TextChange>>", update_highlighting, True)
    wb.bind("<<UpdateAppearance>>", update_highlighting, True)
    



# -*- coding: utf-8 -*-

"""
Adds debugging commands and features. 
"""

import tkinter as tk
from tkinter import ttk
from thonny.common import DebuggerCommand
from thonny.memory import VariablesFrame
from thonny import ast_utils, memory, misc_utils, ui_utils
from thonny.misc_utils import shorten_repr
import ast
from thonny.codeview import CodeView, READ_ONLY_BACKGROUND
from tkinter.messagebox import showinfo, showerror
from thonny.globals import get_workbench, get_runner
from thonny.ui_utils import select_sequence
import tokenize
import logging

_SUSPENDED_FOCUS_BACKGROUND = "#DCEDF2"
_ACTIVE_FOCUS_BACKGROUND = "#F8FC9A"

class Debugger:
    def __init__(self):
        
        self._init_commands()
        
        self._main_frame_visualizer = None
        self._last_progress_message = None
        
        get_workbench().bind("DebuggerProgress", self._handle_debugger_progress, True)
        get_workbench().bind("ToplevelResult", self._handle_toplevel_result, True)
        
        get_workbench().get_view("ShellView").add_command("Debug", 
            get_runner().handle_execute_from_shell)
        
    
    def _init_commands(self):
        get_workbench().add_command("debug", "run", "Debug current script",
            self._cmd_debug_current_script,
            tester=self._cmd_debug_current_script_enabled,
            default_sequence="<Control-F5>",
            group=10,
            image_filename="run.debug_current_script.gif",
            include_in_toolbar=True)
        
        get_workbench().add_command("step_over", "run", "Step over",
            self._cmd_step_over,
            tester=self._cmd_stepping_commands_enabled,
            default_sequence="<F6>",
            group=30,
            image_filename="run.step_over.gif",
            include_in_toolbar=True)
        
        get_workbench().add_command("step_into", "run", "Step into",
            self._cmd_step_into,
            tester=self._cmd_stepping_commands_enabled,
            default_sequence="<F7>",
            group=30,
            image_filename="run.step_into.gif",
            include_in_toolbar=True)
        
        get_workbench().add_command("step_out", "run", "Step out",
            self._cmd_step_out,
            tester=self._cmd_stepping_commands_enabled,
            default_sequence="<F8>",
            group=30,
            image_filename="run.step_out.gif",
            include_in_toolbar=True)
        
        get_workbench().add_command("run_to_cursor", "run", "Run to cursor",
            self._cmd_run_to_cursor,
            tester=self._cmd_run_to_cursor_enabled,
            default_sequence=select_sequence("<Control-F8>", "<Control-F8>"),
            group=30,
            image_filename="run.run_to_cursor.gif",
            include_in_toolbar=False)

    
    def _cmd_debug_current_script(self):
        get_runner().execute_current("Debug")

    def _cmd_debug_current_script_enabled(self):
        return (get_workbench().get_editor_notebook().get_current_editor() is not None
                and get_runner().get_state() == "waiting_toplevel_command"
                and "debug" in get_runner().supported_features())

    def _check_issue_debugger_command(self, command, **kwargs):
        cmd = DebuggerCommand(command=command, **kwargs)
        self._last_debugger_command = cmd
        
        state = get_runner().get_state() 
        if (state == "waiting_debugger_command"
            or getattr(cmd, "automatic", False) and state == "running"):
            logging.debug("_check_issue_debugger_command: %s", cmd)
            
            # tell VM the state we are seeing
            cmd.setdefault (
                frame_id=self._last_progress_message.stack[-1].id,
                state=self._last_progress_message.stack[-1].last_event,
                focus=self._last_progress_message.stack[-1].last_event_focus
            )
            
            get_runner().send_command(cmd)
        else:
            logging.debug("Bad state for sending debugger command " + str(command))


    def _cmd_stepping_commands_enabled(self):
        return get_runner().get_state() == "waiting_debugger_command"
    
    def _cmd_step_into(self):
        self._check_issue_debugger_command("step")
    
    def _cmd_step_over(self):
        # Step over should stop when new statement or expression is selected.
        # At the same time, I need to get value from after_expression event.
        # Therefore I ask backend to stop first after the focus
        # and later I ask it to run to the beginning of new statement/expression.
        
        self._check_issue_debugger_command("exec")
        
    def _cmd_step_out(self):
        self._check_issue_debugger_command("out")

    def _cmd_run_to_cursor(self):
        visualizer = self._get_topmost_selected_visualizer()
        if visualizer:
            assert isinstance(visualizer._text_frame, CodeView)
            code_view = visualizer._text_frame
            selection = code_view.get_selected_range()
            
            target_lineno = visualizer._firstlineno-1 + selection.lineno
            self._check_issue_debugger_command("line",
                                               target_filename=visualizer._filename, 
                                               target_lineno=target_lineno,
                                               )

    def _cmd_run_to_cursor_enabled(self):
        return (self._cmd_stepping_commands_enabled()
                and self._get_topmost_selected_visualizer() is not None
                )

    def _get_topmost_selected_visualizer(self):
        
        visualizer = self._main_frame_visualizer
        if visualizer is None:
            return None
        
        while visualizer._next_frame_visualizer is not None:
            visualizer = visualizer._next_frame_visualizer
        
        topmost_text_widget = visualizer._text
        focused_widget = get_workbench().focus_get()
        
        if focused_widget is None:
            return None
        elif focused_widget == topmost_text_widget:
            return visualizer
        else:
            return None
        

    def _handle_debugger_progress(self, msg):
        self._last_progress_message = msg
        
        if self._should_skip_event(msg):
            self._check_issue_debugger_command("run_to_before", automatic=True)
        else:
            main_frame_id = msg.stack[0].id
            
            # clear obsolete main frame visualizer
            if (self._main_frame_visualizer 
                and self._main_frame_visualizer.get_frame_id() != main_frame_id):
                self._main_frame_visualizer.close()
                self._main_frame_visualizer = None
                
            if not self._main_frame_visualizer:
                self._main_frame_visualizer = MainFrameVisualizer(msg.stack[0])
                
            self._main_frame_visualizer.update_this_and_next_frames(msg)
        
        # advance automatically in some cases
        event = msg.stack[-1].last_event
        args = msg.stack[-1].last_event_args

        if msg.exception:
            showerror("Exception",
                      # Following is clever but noisy 
                      msg.exception_lower_stack_description.lstrip() + 
                      msg.exception["type_name"] 
                      + ": " + msg.exception_msg)
            self._check_issue_debugger_command("step", automatic=True)
            
        elif (event == "after_expression" 
            and "last_child" in args["node_tags"]
            and "child_of_statement" in args["node_tags"]):
            # This means we're done with the expression, so let's speed up a bit.
            self._check_issue_debugger_command("step", automatic=True)
            # Next event will be before_statement_again

                        
            
    
    def _should_skip_event(self, msg):
        frame_info = msg.stack[-1]
        event = frame_info.last_event
        tags = frame_info.last_event_args["node_tags"]
        
        if event == "after_statement":
            return True
        
        # TODO: consult also configuration
        if "call_function" in tags:
            return True
        else:
            return False
    
    def _handle_toplevel_result(self, msg):
        if self._main_frame_visualizer is not None:
            self._main_frame_visualizer.close()
            self._main_frame_visualizer = None    


class FrameVisualizer:
    """
    Is responsible for stepping through statements and updating corresponding UI
    in Editor-s, FunctionCallDialog-s, ModuleDialog-s
    """
    def __init__(self, text_frame, frame_info):
        self._text_frame = text_frame
        self._text = text_frame.text
        self._frame_id = frame_info.id
        self._filename = frame_info.filename
        self._firstlineno = frame_info.firstlineno
        self._source = frame_info.source
        self._expression_box = ExpressionBox(text_frame)
        self._next_frame_visualizer = None
        
        self._text.tag_configure('focus', background=_ACTIVE_FOCUS_BACKGROUND, borderwidth=1, relief=tk.SOLID)
        self._text.tag_configure('exception', background="#FFBFD6")
        self._text.tag_raise("exception", "focus")
        self._text.set_read_only(True)
    
    def close(self):
        if self._next_frame_visualizer:
            self._next_frame_visualizer.close()
            self._next_frame_visualizer = None
            
        self._text.set_read_only(False)
        self._remove_focus_tags()
        self._expression_box.clear_debug_view()
    
    def get_frame_id(self):
        return self._frame_id
    
    def update_this_and_next_frames(self, msg):
        """Must not be used on obsolete frame"""
        
        #debug("State: %s, focus: %s", msg.state, msg.focus)
        
        frame_info, next_frame_info = self._find_this_and_next_frame(msg.stack)
        self._update_this_frame(msg, frame_info)
        
        # clear obsolete next frame visualizer
        if (self._next_frame_visualizer 
            and (not next_frame_info or 
                 self._next_frame_visualizer.get_frame_id() != next_frame_info.id)):
            self._next_frame_visualizer.close()
            self._next_frame_visualizer = None
            
        if next_frame_info and not self._next_frame_visualizer:
            self._next_frame_visualizer = self._create_next_frame_visualizer(next_frame_info)
            
        if self._next_frame_visualizer:
            self._next_frame_visualizer.update_this_and_next_frames(msg)
        
    
    def _remove_focus_tags(self):
        self._text.tag_remove("focus", "0.0", "end")
        self._text.tag_remove("exception", "0.0", "end")
     
    def _update_this_frame(self, msg, frame_info):
        self._frame_info = frame_info
        
        # TODO: if focus is in expression, then find and highlight closest
        # statement
        if "statement" in frame_info.last_event:
            self._remove_focus_tags()
            self._tag_range(frame_info.last_event_focus, "focus", True)
            if msg.exception is not None:
                self._tag_range(frame_info.last_event_focus, "exception", True)
                
            self._text.tag_configure('focus', background=_ACTIVE_FOCUS_BACKGROUND, borderwidth=1, relief=tk.SOLID)
        else:
            self._text.tag_configure('focus', background=READ_ONLY_BACKGROUND, borderwidth=1, relief=tk.SOLID)
            
        self._expression_box.update_expression(msg, frame_info)

    def _find_this_and_next_frame(self, stack):
        for i in range(len(stack)):
            if stack[i].id == self._frame_id:
                if i == len(stack)-1: # last frame
                    return stack[i], None
                else:
                    return stack[i], stack[i+1]
                    
        else:
            raise AssertionError("Frame doesn't exist anymore")
        
    
    def _tag_range(self, text_range, tag, see=False):
        first_line, first_col, last_line = self._get_text_range_block(text_range)
        
        for lineno in range(first_line, last_line+1):
            self._text.tag_add(tag,
                              "%d.%d" % (lineno, first_col),
                              "%d.0" % (lineno+1))
            
        self._text.update_idletasks()
        self._text.see("%d.0" % (first_line))

        if last_line - first_line < 3:
            # if it's safe to assume that whole code fits into screen
            # then scroll it down a bit so that expression view doesn't hide behind
            # lower edge of the editor
            self._text.update_idletasks()
            self._text.see("%d.0" % (first_line+3))
            
    def _get_text_range_block(self, text_range):
        first_line = text_range.lineno - self._frame_info.firstlineno + 1
        last_line = text_range.end_lineno - self._frame_info.firstlineno + 1
        first_line_content = self._text.get("%d.0" % first_line, "%d.end" % first_line)
        if first_line_content.strip().startswith("elif "):
            first_col = first_line_content.find("elif ")
        else:
            first_col = text_range.col_offset
        
        return (first_line, first_col, last_line)
    
            
    
    def _create_next_frame_visualizer(self, next_frame_info):
        if next_frame_info.code_name == "<module>":
            return ModuleLoadDialog(self._text, next_frame_info)
        else:
            dialog = FunctionCallDialog(self._text.master, next_frame_info)
            
            if self._expression_box.winfo_ismapped():
                dialog.title(self._expression_box.get_focused_text())
            else:
                dialog.title("Function call at " + hex(self._frame_id))
                 
            return dialog
     


class MainFrameVisualizer(FrameVisualizer):
    """
    Takes care of stepping in the main module
    """
    def __init__(self, frame_info):
        editor = get_workbench().get_editor_notebook().show_file(frame_info.filename)
        FrameVisualizer.__init__(self, editor.get_code_view(), frame_info)
        

class CallFrameVisualizer(FrameVisualizer):
    def __init__(self, text_frame, frame_id):
        self._dialog = FunctionCallDialog(text_frame)
        FrameVisualizer.__init__(self, self._dialog.get_code_view(), frame_id)
        
    def close(self):
        super().close()
        self._dialog.destroy()

class ExpressionBox(tk.Text):
    def __init__(self, codeview):
        tk.Text.__init__(self, codeview.winfo_toplevel(), #codeview.text,
                         height=1,
                         width=1,
                         relief=tk.RAISED,
                         background="#DCEDF2",
                         borderwidth=1,
                         highlightthickness=0,
                         padx=7,
                         pady=7,
                         wrap=tk.NONE,
                         font=get_workbench().get_font("EditorFont"))
        self._codeview = codeview
        
        self._main_range = None
        self._last_focus = None
        
        self.tag_config("value", foreground="Blue")
        self.tag_configure('before', background="#F8FC9A", borderwidth=1, relief=tk.SOLID)
        self.tag_configure('after', background="#D7EDD3", borderwidth=1, relief=tk.FLAT)
        self.tag_configure('exception', background="#FFBFD6", borderwidth=1, relief=tk.SOLID)
        self.tag_raise("exception", "before")
        self.tag_raise("exception", "after")
        
        
    def update_expression(self, msg, frame_info):
        focus = frame_info.last_event_focus
        event = frame_info.last_event
        
        if event in ("before_expression", "before_expression_again"):
            # (re)load stuff
            if self._main_range is None or focus.not_smaller_eq_in(self._main_range):
                self._load_expression(frame_info.filename, focus)
                self._update_position(focus)
                self._update_size()
                
            self._highlight_range(focus, event, msg.exception)
            
        
        elif event == "after_expression":
            logging.debug("EV: after_expression %s", msg)
            
            self.tag_configure('after', background="#BBEDB2", borderwidth=1, relief=tk.FLAT)
            start_mark = self._get_mark_name(focus.lineno, focus.col_offset)
            end_mark = self._get_mark_name(focus.end_lineno, focus.end_col_offset)
            
            assert hasattr(msg, "value")
            logging.debug("EV: replacing expression with value")

            original_focus_text = self.get(start_mark, end_mark)
            self.delete(start_mark, end_mark)
            
            id_str = memory.format_object_id(msg.value["id"])
            if get_workbench().in_heap_mode():
                value_str = id_str
            elif "StringLiteral" in frame_info.last_event_args["node_tags"]:
                # No need to show Python replacing double quotes with single quotes
                value_str = original_focus_text
            else:
                value_str = shorten_repr(msg.value["repr"], 100)
            
            object_tag = "object_" + str(msg.value["id"])
            self.insert(start_mark, value_str, ('value', 'after', object_tag))
            if misc_utils.running_on_mac_os():
                sequence = "<Command-Button-1>"
            else:
                sequence = "<Control-Button-1>"
            self.tag_bind(object_tag, sequence,
                          lambda _: get_workbench().event_generate("ObjectSelect", object_id=msg.value["id"]))
                
            self._update_size()
                
            
                
        elif (event == "before_statement_again"
              and self._main_range is not None # TODO: shouldn't need this 
              and self._main_range.is_smaller_eq_in(focus)):
            # we're at final stage of executing parent statement 
            # (eg. assignment after the LHS has been evaluated)
            # don't close yet
            self.tag_configure('after', background="#DCEDF2", borderwidth=1, relief=tk.FLAT)
        
        elif event == "exception":
            "TODO:"   
        
        else:
            # hide and clear on non-expression events
            self.clear_debug_view()

        self._last_focus = focus
        
        
    def get_focused_text(self):
        if self._last_focus: 
            start_mark = self._get_mark_name(self._last_focus.lineno, self._last_focus.col_offset)
            end_mark = self._get_mark_name(self._last_focus.end_lineno, self._last_focus.end_col_offset)
            return self.get(start_mark, end_mark)
        else:
            return ""
      
    def clear_debug_view(self):
        self.place_forget()
        self._main_range = None
        self._last_focus = None
        for tag in self.tag_names():
            self.tag_remove(tag, "1.0", "end")
            
        for mark in self.mark_names():
            self.mark_unset(mark)
            
    
    def _load_expression(self, filename, text_range):
        with tokenize.open(filename) as fp:
            whole_source = fp.read()
            
        root = ast_utils.parse_source(whole_source, filename)
        self._main_range = text_range
        assert self._main_range is not None
        main_node = ast_utils.find_expression(root, text_range)
        
        source = ast_utils.extract_text_range(whole_source, text_range)
        logging.debug("EV.load_exp: %s", (text_range, main_node, source))
        
        self.delete("1.0", "end")
        self.insert("1.0", source)
        
        # create node marks
        def _create_index(lineno, col_offset):
            local_lineno = lineno - main_node.lineno + 1
            if lineno == main_node.lineno:
                local_col_offset = col_offset - main_node.col_offset
            else:
                local_col_offset = col_offset
                
            return str(local_lineno) + "." + str(local_col_offset)

        for node in ast.walk(main_node):
            if "lineno" in node._attributes:
                index1 = _create_index(node.lineno, node.col_offset)
                index2 = _create_index(node.end_lineno, node.end_col_offset)
                
                start_mark = self._get_mark_name(node.lineno, node.col_offset) 
                if not start_mark in self.mark_names():
                    self.mark_set(start_mark, index1)
                    self.mark_gravity(start_mark, tk.LEFT)
                
                end_mark = self._get_mark_name(node.end_lineno, node.end_col_offset) 
                if not end_mark in self.mark_names():
                    self.mark_set(end_mark, index2)
                    self.mark_gravity(end_mark, tk.RIGHT)
                    
    def _get_mark_name(self, lineno, col_offset):
        return str(lineno) + "_" + str(col_offset)
    
    def _get_tag_name(self, node_or_text_range):
        return (str(node_or_text_range.lineno)
                + "_" + str(node_or_text_range.col_offset)
                + "_" + str(node_or_text_range.end_lineno)
                + "_" + str(node_or_text_range.end_col_offset))
    
    def _highlight_range(self, text_range, state, exception):
        logging.debug("EV._highlight_range: %s", text_range)
        self.tag_remove("after", "1.0", "end")
        self.tag_remove("before", "1.0", "end")
        self.tag_remove("exception", "1.0", "end")
        
        if state.startswith("after"):
            tag = "after"
        elif state.startswith("before"):
            tag = "before"
        else:
            return
        
        start_index = self._get_mark_name(text_range.lineno, text_range.col_offset)
        end_index = self._get_mark_name(text_range.end_lineno, text_range.end_col_offset) 
        self.tag_add(tag, start_index, end_index)
        
        if exception:
            self.tag_add("exception", start_index, end_index)
            
    def _update_position(self, text_range):
        self._codeview.update_idletasks()
        text_line_number = text_range.lineno - self._codeview._first_line_number + 1
        bbox = self._codeview.text.bbox(str(text_line_number) + "." + str(text_range.col_offset))
        
        if isinstance(bbox, tuple):
            x = bbox[0] - self._codeview.text.cget("padx") + 6 
            y = bbox[1] - self._codeview.text.cget("pady") - 6
        else:
            x = 30
            y = 30
            
        widget = self._codeview.text
        while widget != self.winfo_toplevel():
            x += widget.winfo_x()
            y += widget.winfo_y()
            widget = widget.master
            
        self.place(x=x, y=y, anchor=tk.NW)
        self.update_idletasks()

    
    def _update_size(self):
        content = self.get("1.0", tk.END)
        lines = content.splitlines()
        self["height"] = len(lines)
        self["width"] = max(map(len, lines))
    

class FrameDialog(tk.Toplevel, FrameVisualizer):
    def __init__(self, master, frame_info):
        tk.Toplevel.__init__(self, master)
        
        self.transient(master)
        if misc_utils.running_on_windows():
            self.wm_attributes('-toolwindow', 1)
        
        
        # TODO: take size from prefs
        editor_notebook = get_workbench().get_editor_notebook()
        if master.winfo_toplevel() == get_workbench():
            position_reference = editor_notebook
        else:
            # align to previous frame
            position_reference = master.winfo_toplevel()
            
        self.geometry("{}x{}+{}+{}".format(editor_notebook.winfo_width(),
                                           editor_notebook.winfo_height()-20,
                                           position_reference.winfo_rootx(),
                                           position_reference.winfo_rooty()))
        self.protocol("WM_DELETE_WINDOW", self._on_close)
        
        self._init_layout_widgets(master, frame_info)
        FrameVisualizer.__init__(self, self._text_frame, frame_info)
        
        self._load_code(frame_info)
        self._text_frame.text.focus()
    
    def _init_layout_widgets(self, master, frame_info):
        self.main_frame= ttk.Frame(self) # just a backgroud behind padding of main_pw, without this OS X leaves white border
        self.main_frame.grid(sticky=tk.NSEW)        
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.main_pw = ui_utils.AutomaticPanedWindow(self.main_frame, orient=tk.VERTICAL)
        self.main_pw.grid(sticky=tk.NSEW, padx=10, pady=10)
        self.main_frame.rowconfigure(0, weight=1)
        self.main_frame.columnconfigure(0, weight=1)
        
        self._code_book = ttk.Notebook(self.main_pw)
        self._text_frame = CodeView(self._code_book, 
                                      first_line_number=frame_info.firstlineno,
                                      font=get_workbench().get_font("EditorFont"))
        self._code_book.add(self._text_frame, text="Source")
        self.main_pw.add(self._code_book, minsize=100)
        
    
    def _load_code(self, frame_info):
        self._text_frame.set_content(frame_info.source)
    
    def _update_this_frame(self, msg, frame_info):
        FrameVisualizer._update_this_frame(self, msg, frame_info)
    
    def _on_close(self):
        showinfo("Can't close yet", 'Use "Stop" command if you want to cancel debugging')
    
    
    def close(self):
        FrameVisualizer.close(self)
        self.destroy()

class FunctionCallDialog(FrameDialog):
    def __init__(self, master, frame_info):
        FrameDialog.__init__(self, master, frame_info)
    
    def _init_layout_widgets(self, master, frame_info):
        FrameDialog._init_layout_widgets(self, master, frame_info)
        self._locals_book = ttk.Notebook(self.main_pw)
        self._locals_frame = VariablesFrame(self._locals_book)
        self._locals_book.add(self._locals_frame, text="Local variables")
        self.main_pw.add(self._locals_book, minsize=100)

    def _load_code(self, frame_info):
        FrameDialog._load_code(self, frame_info)
        
        if hasattr(frame_info, "function"):
            function_label = frame_info.function["repr"]
        else:
            function_label = frame_info.code_name
        
        # change tab label
        self._code_book.tab(self._text_frame, text=function_label)
    
    def _update_this_frame(self, msg, frame_info):
        FrameDialog._update_this_frame(self, msg, frame_info)
        self._locals_frame.update_variables(frame_info.locals)

        
class ModuleLoadDialog(FrameDialog):
    def __init__(self, text_frame, frame_info):
        FrameDialog.__init__(self, text_frame)
    
    
    
def load_plugin():
    Debugger()
    
        


# -*- coding: utf-8 -*-

from thonny.memory import VariablesFrame
from thonny.globals import get_workbench, get_runner
from thonny.common import InlineCommand

class GlobalsView(VariablesFrame):
    def __init__(self, master):
        VariablesFrame.__init__(self, master)
        
        get_workbench().bind("Globals", self._handle_globals_event, True)
        get_workbench().bind("BackendRestart", lambda e=None: self._clear_tree(), True)
        get_workbench().bind("DebuggerProgress", self._request_globals, True)
        get_workbench().bind("ToplevelResult", self._request_globals, True)
    
    def before_show(self):
        self._request_globals(even_when_hidden=True)
    
    def _handle_globals_event(self, event):
        # TODO: handle other modules as well
        self.update_variables(event.globals)
    
    def _request_globals(self, event=None, even_when_hidden=False):
        if self.winfo_ismapped() or even_when_hidden:
            # TODO: module_name
            get_runner().send_command(InlineCommand("get_globals", module_name="__main__"))
    

def load_plugin():
    get_workbench().add_view(GlobalsView, "Variables", "ne")


# -*- coding: utf-8 -*-

import tkinter as tk
import os.path
import ast
from configparser import ConfigParser
import configparser
from logging import exception
from tkinter import messagebox

def try_load_configuration(filename):
    try: 
        return ConfigurationManager(filename)
    except configparser.Error:
        if (os.path.exists(filename) 
            and messagebox.askyesno("Problem", 
                "Thonny's configuration file can't be read. It may be corrupt.\n\n"
                + "Do you want to discard the file and open Thonny with default settings?")):
            os.replace(filename, filename + "_corrupt")
            # For some reasons Thonny styles are not loaded properly once messagebox has been shown before main window (At least Windows Py 3.5)
            raise SystemExit("Configuration file has been discarded. Please restart Thonny!")
        else:
            raise
    

class ConfigurationManager:
    def __init__(self, filename):
        self._ini = ConfigParser()
        self._filename = filename
        self._defaults = {}
        self._variables = {} # Tk variables
        
        if os.path.exists(self._filename):
            with open(self._filename, 'r', encoding="UTF-8") as fp: 
                self._ini.read_file(fp)

        #print(prefs_filename, self.sections())
    
    def get_option(self, name, secondary_default=None):
        section, option = self._parse_name(name)
        name = section + "." + option
        
        # variable may have more recent value
        if name in self._variables:
            return self._variables[name].get()
            
        try:
            val = self._ini.get(section, option)
            try:
                return ast.literal_eval(val)
            except:
                return val
        except:
            if name in self._defaults:
                return self._defaults[name]
            else:
                return secondary_default
            
    
    def has_option(self, name):
        return name in self._defaults
    
    def set_option(self, name, value):
        section, option = self._parse_name(name)
        name = section + "." + option
        if not self._ini.has_section(section):
            self._ini.add_section(section)
        
        if isinstance(value, str):
            self._ini.set(section, option, value)
        else:
            self._ini.set(section, option, repr(value))
        
        # update variable
        if name in self._variables:
            self._variables[name].set(value)
    
    def set_default(self, name, primary_default_value):
        section, option = self._parse_name(name)
        name = section + "." + option
        self._defaults[name] = primary_default_value

    def get_variable(self, name):
        section, option = self._parse_name(name)
        name = section + "." + option
        
        if name in self._variables:
            return self._variables[name]
        else:
            value = self.get_option(name)
            if isinstance(value, bool):
                var = tk.BooleanVar(value=value)
            elif isinstance(value, int):
                var = tk.IntVar(value=value)
            elif isinstance(value, str):
                var = tk.StringVar(value=value)
            else:
                raise KeyError("Can't create Tk Variable for " + name)
            self._variables[name] = var
            return var
    
    def save(self):
        # save all tk variables
        for name in self._variables:
            self.set_option(name, self._variables[name].get())
            
        # store
        if not os.path.exists(self._filename):
            os.makedirs(os.path.dirname(self._filename), mode=0o700, exist_ok=True)

        # Normal saving occasionally creates corrupted file:
        # https://bitbucket.org/plas/thonny/issues/167/configuration-file-occasionally-gets
        # Now I'm saving the configuration to a temp file 
        # and if the save is successful, I replace configuration file with it
        temp_filename = self._filename + ".temp"             
        with open(temp_filename, 'w', encoding="UTF-8") as fp:
            self._ini.write(fp)
            
        try:
            ConfigurationManager(temp_filename)
            # temp file was created successfully
            os.chmod(temp_filename, 0o600)
            os.replace(temp_filename, self._filename)
            os.chmod(self._filename, 0o600)
        except:
            exception("Could not save configuration file. Reverting to previous file.")
        

    def _parse_name(self, name):
        if "." in name:
            return name.split(".", 1)
        else:
            return "general", name 
